#!/usr/bin/env python3

# ----------------------------------------------------------------------------------------------------------------------
# pfemame.py

# Here's my docstring
"""
Python Front End for the MAME emulation system

Author Gareth Finch

This is a front-end for MAME written in Python and using the wx graphical toolkit.  It is cross-platform and has
been tested in Windows (8, 10, 11), Ubuntu Linux, WSL2-Ubuntu, and MacOS.
Minimum requirements are;
 python 3.10 or later (64 bit)
 wxPython 4.2.0 or later
 pypubsub
 python-docx
 python-vlc (Although it will work without this and just fall back to MediaCtrl)
 vlc 64 bit if this is to be used for video playback
 cx_Freeze 6.0 or later if freezing the application for distribution for Windows or Linux
 py2app for freezing the application for distribution for macOS
 MAME 0.206 or later

For first time users or people wanting to build from this source, please read the manual from front to back as it
answers a lot of questions.
"""

# Import python & wxPython modules
import wx
import wx.adv
import glob
import wx.media
import wx.lib.scrolledpanel
import wx.lib.agw.pygauge as gauge
import os
import subprocess
import languagepicker
# import wx.lib.agw.ultimatelistctrl as ulc
import ultimatelistctrl as ulc
import sys
import shutil
from random import randint
import shelve
import urllib
from urllib.request import Request, urlopen
import webbrowser
import multiprocessing
import time
import datetime
from threading import Thread
import locale
import wx.lib.agw.flatnotebook as flatnotebook
import ssl
from typing import Union
import argparse

try:
    from pubsub import pub
except ModuleNotFoundError:
    print('ERROR!! - You need to install pypubsub.  Use pip install pypubsub, '
          'or pip3 install pypubsub if required')
    exit()
try:
    import docx
except ModuleNotFoundError:
    print('ERROR!! - You need to install python-docx.  Use pip install python-docx, '
          'or pip3 install python-docx if required')
    exit()
try:
    import vlc
    force_disable_vlc = False
except ModuleNotFoundError:
    print('ERROR!! - You need to install python-vlc to use the  VLC backend.  Use pip install python-vlc, '
          'or pip3 install python-vlc if required.  VLC Support will be disabled.')
    force_disable_vlc = True
except OSError:
    print('ERROR:  Unable to load required VLC library.  The application needs to find libvlc.dll and '
          'can\'t.  You need to ensure you have the 64bit version of VLC installed.  VLC Support will be disabled')
    force_disable_vlc = True

# Import my own modules
import iniFile
import filepaths
import rominfo
import preferences
import tileview
import checkall
import rungame
import functions
import scrolledwindow
import database_build
import xmlinfo
import aboutbox
import romRating
import busywindow
import fullscreenplay
import popupimage
import alternate_emulators

# Define Constants here
__version__ = '2.29.4'
__author__ = 'Gareth Finch'
__copyright__ = u'(c)Gareth Finch 2014 - 2023'
__credits__ = ['Dev Player']
__license__ = 'GPL'
__maintainer__ = 'Gareth Finch'
__email__ = 'pfemame@gmail.com'
__status__ = 'Production'
__developer__ = False

# Define the mylist index locations for the table data
index_rom_name = 0
index_rom_file_name = 1
index_system_program_file_name = 2
index_exists = 3
index_rating = 4
index_number_runs = 5
index_favorite = 6
index_category = 7
index_system_short_name = 8
index_system_folder_name = 9
index_path_media_name = 9
index_year = 10
index_manufacturer = 11

# Define column widths
rom_name_col_size = 220
file_name_col_size = 60
system_file_name_col_size = 100
exists_col_size = 50
short_name_col_size = 100
rating_col_size = 100
number_of_runs_col_size = 50
favorite_col_size = 50
category_col_size = 80
path_media_name_col_size = 120
year_col_size = 50
manufacturer_col_size = 120

# Define the number of columns in the main view list (note that some of these are hidden).  This value needs to be
# updated if I change the amount of data that I want to store / display.
display_columns = 12

# Other misc constants
systems_list = ['Home Computer',
                'Game Console',
                'Business Computer',
                'Handheld Game',
                'Handheld Game Console',
                'Computer',
                'Handheld',
                'System',
                'Misc.',
                'Music',
                'Computer Graphic Workstation',
                'Utilities']
messageDelimiter = '---------------------------------------------------------------------------------------------' \
                   '-------------------------------------------------------------------------\n'

files_to_copy = ['categories.dat',
                 'favorites.dat',
                 'ratings.dat',
                 'runs.dat',
                 'pfemame.ini',
                 'non_mame_emulators.xml']


# This is the mainloop class and is the program entry point
class MainFrame(wx.Frame):
    # Create class variables here that can be accessed in other classes
    mylist_filtered = []
    image_file_name = ''

    def __init__(self):
        wx.Frame.__init__(self, None, -1, f'pfeMAME {__version__}', pos=(0, 0), size=(-1, -1),
                          # style=wx.DEFAULT_FRAME_STYLE ^ wx.RESIZE_BORDER ^ wx.MAXIMIZE_BOX)
                          style = wx.DEFAULT_FRAME_STYLE)

        # Get the application language
        global _
        _ = functions.set_up_language()  # Set up the application language
        self.language = iniFile.read_ini_file('preferences', 'language', 'english')

        # --------------------------------------------------------------------------------------------------------------
        # Look for and process command line options.
        self.run_specified_rom = False
        forced_restart = ''
        parser = argparse.ArgumentParser()
        parser.add_argument('-d', '--developer', help='Developer mode', action='store_true')
        parser.add_argument('-novlc', '--novlc', help='Disable VLC backend support', action='store_true')
        parser.add_argument('-forcevid', '--force_video', help='Force video playback', action='store_true')
        parser.add_argument('-novid', '--no_video', help='Disable video playback', action='store_true')
        parser.add_argument('-x', '--force_restart', help='Force a restart (really just for testing purposes)',
                            action='store_true')
        parser.add_argument('-v', '--version', help='Display the pfemame version', action='store_true')
        parser.add_argument('-r', '--run', help='Run the ROM specified in romfilename', type=str,
                            metavar='{rom file name}')
        parser.add_argument('-l', '--language', help='Set the language', type=str,
                            choices=['english', 'spanish', 'german', 'chinese', 'japanese', 'russian'])
        args = parser.parse_args()

        # Check if we are trying to force video playback (This will override the check for X11 on Linux
        self.force_video = False
        if args.force_video:
            self.force_video = True
            print('Forcing Video playback support')
        # Check for vlc disable
        self.force_disable_vlc = force_disable_vlc
        if args.novlc:
            self.force_disable_vlc = True
            print('Disabling VLC playback support')
        self.development = True if (args.developer or __developer__ is True) else False
        # Check for developer mode
        self.development = True if (args.developer or __developer__ is True) else False
        # Check for video disable
        self.video_disable = True if args.no_video else False
        # Check for forced restart
        if args.force_restart:
            forced_restart = (_('pfeMAME application restarted'))
            print('forced')
        # Check for version
        if args.version:
            print(f'pfeMAME version {__version__}')
            exit()
        # Check for run rom
        if args.run:
            self.specified_rom_name = args.run
            self.run_specified_rom = True
        # Check for language
        if args.language:
            self.language = args.language.lower()
            iniFile.set_ini_file_variables([('preferences', 'language', self.language)])
            _ = functions.set_up_language()  # Set up the application language
            self.Refresh()
            print(f'Language has been set to {self.language}')

        # --------------------------------------------------------------------------------------------------------------
        # Create the empty development info startup log list. This is appended after each main block of code is run
        # so that I can see execution times for debug purposes.
        self.development_startup_info = []

        # --------------------------------------------------------------------------------------------------------------
        # Create a snapshot of the current time so that we can determine the application startup time
        # as well as the total application run time for the application ended log.
        self.startup_process_start_time = time.time()
        self.debug_start_time = time.time()

        # --------------------------------------------------------------------------------------------------------------
        # Ensure that locale is set to the system default.  This is needed as the datetime functions fail sometimes
        # without it.
        locale.setlocale(locale.LC_ALL, '')

        # --------------------------------------------------------------------------------------------------------------
        # This makes sure that the multiprocessor calls are supported in frozen applications
        multiprocessing.freeze_support()

        # --------------------------------------------------------------------------------------------------------------
        # Display the wait information while the application is started up
        self.start_up_complete = False  # Used by the single click event handler

        # --------------------------------------------------------------------------------------------------------------
        # Bind the closing event, so we can close gracefully
        self.Bind(wx.EVT_CLOSE, self.on_close)

        # --------------------------------------------------------------------------------------------------------------
        # Define the application icon
        self.SetIcon(wx.Icon(os.path.join(path_to_me, 'images', 'pfemame.ico'), wx.BITMAP_TYPE_ICO))

        # --------------------------------------------------------------------------------------------------------------
        # Force the system current working directory to the location of this file.  This ensures that all path and file
        # checks later find the files they are looking for.  Found this was an issue under Visual Studio Code when
        # running pfeMAME from a different directory.
        if getattr(sys, 'frozen', False):
            # The application is frozen
            data_directory = os.path.dirname(sys.executable)
        else:
            # The application is not frozen
            data_directory = os.path.dirname(os.path.abspath(__file__))
        os.chdir(data_directory)

        # --------------------------------------------------------------------------------------------------------------
        # Try to work out the platform
        # This is needed to sort out some platform specific differences
        self.system = functions.platform()

        # --------------------------------------------------------------------------------------------------------------
        # Create a pubsub receiver for the thread used by rungame
        # These allow messages to be sent back to the main screen from a thread
        pub.subscribe(self.update_display_messages, 'update')  # This is for message window updates
        pub.subscribe(self.update_logfile, 'log')  # This is for log file
        pub.subscribe(self.update_display_popups, 'popup')  # This is for popup messages
        pub.subscribe(self.update_display_number_of_runs, 'number_of_runs')  # This is for number of runs updating
        pub.subscribe(self.enable_alert_timer, 'enable_alert_timer')  # Start the alert timer for alert colors strobing
        pub.subscribe(self.update_status, 'status')  # This is for status bar updates
        pub.subscribe(self.setup_ini_variables, 'setup_ini')  # This is for status bar updates
        pub.subscribe(self.on_force_restart, 'force_restart')  # This is for forced application restarts

        # --------------------------------------------------------------------------------------------------------------
        # Set the shell type to use for subprocess depending upon the operating system type
        self.shell_to_use = (True and self.system == 'windows')

        # --------------------------------------------------------------------------------------------------------------
        # Get the path to this file
        self.path_to_me = functions.path_to_me()  # Get the absolute path to me

        # --------------------------------------------------------------------------------------------------------------
        # Load the alternate emulation engine data
        (self.alternate_emulation_engines,
         self.alternate_emulation_engines_error) = functions.load_other_emulators_detail(self.path_to_me)

        # --------------------------------------------------------------------------------------------------------------
        # Set up the log file
        self.my_logger = functions.setup_logging(self.path_to_me)
        self.my_logger.info(_('Application Started'))  # Write to the log file
        self.my_logger.info(f'Platform: {sys.platform}')  # Write to the log file
        self.my_logger.info(f'Python Version: {sys.version}')  # Write to the log file
        self.my_logger.info(f'WX Version: {wx.version()}')  # Write to the log file

        # --------------------------------------------------------------------------------------------------------------
        # Determine which video playback media player to use
        self.use_vlc, self.media_player_available = functions.media_player_to_use(self.system, self.force_disable_vlc)
        if not self.media_player_available:
            self.video_disable = True

        # --------------------------------------------------------------------------------------------------------------
        # If the OS is Linux, check if we are using X11 or wayland.  If wayland then disable video playback as
        # there are serious issues with media ctrl and vlc under wayland.
        if self.system == 'linux' and os.environ['XDG_SESSION_TYPE'] == 'wayland' and not self.force_video:
            self.my_logger.info(_('Wayland is display manager - Video playback disabled'))  # Write to the log file
            self.video_disable = True

        # --------------------------------------------------------------------------------------------------------------
        # Get todays date / time - used for age of cache file checking later on
        self.today = datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%d %H:%M:%S')

        # ------------------------------------------------------------------------------------------------------------------
        # Get variables from the mame.ini and pfemame.ini files
        self.setup_ini_variables()

        # --------------------------------------------------------------------------------------------------------------
        # Set some other default variables
        self.mylist = []
        self.history_data = []
        self.mameinfo_data = []
        self.clearSearch = True
        self.video_to_play = ''
        self.availableRoms = 0
        self.alert_colors_decrement_value = 16
        self.alert_colors_value = 240
        self.tileview_running = False
        self.joystick_block = False
        self.inactivity_timer_block = False
        self.image_options_index = 0  # Start with Snapshots displayed
        self.overlay_image = wx.Bitmap(os.path.join(self.path_to_me, 'images', 'lg2.png'), wx.BITMAP_TYPE_ANY)
        self.overlay_image_size = self.overlay_image.GetSize()
        self.close_selected = False
        self.vlc_playing = False
        self.rom_name_col_size = rom_name_col_size
        self.file_name_col_size = file_name_col_size
        self.system_file_name_col_size = system_file_name_col_size

        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Initial variables set'))

        # --------------------------------------------------------------------------------------------------------------
        # Set up the alert timer
        self.alert_timer = wx.Timer(self, id=9991)
        self.Bind(wx.EVT_TIMER, self.on_timer_event, self.alert_timer)

        # --------------------------------------------------------------------------------------------------------------
        # Set up the inactivity timer and start it.  This is used for the screensaver.
        self.inactivity_timer = wx.Timer(self, id=9992)
        self.Bind(wx.EVT_TIMER, self.on_timer_event, self.inactivity_timer)
        self.inactivity_timer_start()
 
        # --------------------------------------------------------------------------------------------------------------
        # Define the menu bar items and event bindings
        menu_bar = wx.MenuBar()

        # Redirect functions for the MenuBar help text.  This sends it to my overlay statusbar rather than the 
        # system default statusbar.
        def on_help_string(event):
            try:
                event_id = event.GetId()
                self.statusbar_panel.set_status(menu_bar.GetHelpString(event_id))
            except wx._core.wxAssertionError:
                self.statusbar_panel.set_status(_('Ready'))

        def on_remove_text(_event):
            self.statusbar_panel.set_status(_('Ready'))

        file_menu = wx.Menu()
        edit_menu = wx.Menu()
        view_menu = wx.Menu()
        help_menu = wx.Menu()

        temp = [(file_menu,
                 os.path.join(path_to_me, 'images', 'play.png'),
                 _('&Play ROM'),
                 _('Play the selected ROM'),
                 self.on_double_click
                 ),
                (file_menu,
                 os.path.join(path_to_me, 'images', 'play_random.png'),
                 _('Play Random ROM'),
                 _('Play a random ROM (any ROM that exists)'),
                 self.on_random
                 ),
                (file_menu,
                 os.path.join(path_to_me, 'images', 'verifyrom.png'),
                 _('&Verify ROMs'),
                 _('Verify the ROMs for the selected ROM'),
                 self.on_verify_roms
                 ),
                (file_menu,
                 '',
                 _('V&iew Media Types'),
                 _('View the types of media that this system is capable of running'),
                 self.on_view_media_types
                 ),
                (file_menu,
                 os.path.join(path_to_me, 'images', 'info.png'),
                 _('&ROM Info'),
                 _('Display detailed ROM information'),
                 self.on_rom_info
                 ),
                (file_menu,
                 '',
                 _('Reset INI File'),
                 _('Reset the INI configuration file for this program'),
                 self.on_ini_file_reset
                 ),
                (file_menu, 'separator'),
                (file_menu,
                 '',
                 _('&Check All ROMs'),
                 _('Carry out a verification on all ROMs listed as existing in your collection'),
                 self.on_check_all
                 ),
                (file_menu, 'separator'),
                (file_menu,
                 '',
                 _('Record ROM Input'),
                 _('Record the control inputs for the selected ROM'),
                 self.on_input_record
                 ),
                (file_menu,
                 '',
                 _('Playback ROM Input'),
                 _('Play the selected ROM using previously recorded inputs'),
                 self.on_input_playback
                 ),
                (file_menu, 'separator'),
                (file_menu,
                 '',
                 _('&Backup All Settings'),
                 _('Backup preferences and DAT files'),
                 self.on_backup
                 ),
                (file_menu,
                 '',
                 _('&Load Settings From Backup'),
                 _('Restore previously backed up preferences and DAT files'),
                 self.on_restore
                 ),
                (file_menu, 'separator'),
                (file_menu,
                 os.path.join(path_to_me, 'images', 'exit.png'),
                 '&Quit\tCtrl+Q',
                 _('Quit the Application'),
                 self.on_close
                 ),
                (edit_menu,
                 '',
                 _('&Add to Favorites'),
                 _('Add the selected ROM to favorites'),
                 self.on_add_favorite
                 ),
                (edit_menu,
                 '',
                 _('&ROM Rating'),
                 _('Set the star rating for the ROM'),
                 self.on_set_rating
                 ),
                (edit_menu,
                 os.path.join(path_to_me, 'images', 'filepaths.png'),
                 _('File Paths'),
                 _('Edit the file and folder paths'),
                 self.on_file_paths
                 ),
                (edit_menu,
                 '',
                 _('Setup Alternate Emulators'),
                 _('Set up alternate emulator engines'),
                 self.on_alternate_emulators
                 ),
                (edit_menu,
                 os.path.join(path_to_me, 'images', 'search.png'),
                 '&Find ROM\tCtrl+F',
                 _('Find a ROM Name, File Name or System File Name'),
                 self.on_find
                 ),
                (edit_menu,
                 '',
                 _('Reset pfeMAME INI file'),
                 _('Reset the pfeMAME INI file back to default'),
                 self.on_reset_pfemame_ini
                 ),
                (edit_menu,
                 '',
                 _('Messages - Clear'),
                 _('Clear the messages window'),
                 self.on_clear_messages
                 ),
                (edit_menu,
                 '',
                 _('Messages - Copy to Clipboard'),
                 _('Copy the messages window to the clipboard'),
                 self.on_copy_messages
                 ),
                (edit_menu,
                 '',
                 _('Reset Spacing'),
                 _('Reset the ROMs list column spacing to default'),
                 self.on_reset_spacing
                 ),
                (view_menu,
                 '',
                 _('&Full screen video player\tSHIFT-V'),
                 _('Start the full screen video player (AKA screensaver)'),
                 self.on_fullscreen_video_player
                 ),
                (view_menu,
                 '',
                 _('&View pfeMAME Log File'),
                 _('View the automatically created pfeMAME log file'),
                 self.on_logfile_display
                 ),
                (view_menu,
                 os.path.join(path_to_me, 'images', 'refresh.png'),
                 _('&Refresh ROMs Database\tF5'),
                 _('Reload and refresh the full ROMs database'),
                 self.on_database_refresh
                 ),
                (view_menu,
                 '',
                 _('&Switch to Tile View'),
                 _('Show the ROMs in a tile view'),
                 self.on_tile_view
                 ),
                (help_menu,
                 '',
                 _('&Controls'),
                 _('MAME controls'),
                 self.on_controls_box_display
                 ),
                (help_menu,
                 '',
                 _('&Manual'),
                 _('pfeMAME manual (pdf copy)'),
                 self.on_manual_display
                 ),
                (help_menu,
                 '',
                 _('C&heck for updates'),
                 _('Check for software updates to pfeMAME'),
                 self.check_for_updates
                 ),
                (help_menu,
                '',
                 _('pfeMAME web page'),
                 _('Open the pfeMAME sourceforge web page'),
                 self.open_web_page
                 ),
                (help_menu,
                '',
                 _('View Change Log'),
                 _('Open the pfeMAME changelog file'),
                 self.on_changelog_display
                 ),
                (help_menu, 'separator'),
                (help_menu,
                 wx.ArtProvider.GetBitmap(wx.ART_QUESTION, wx.ART_BUTTON, (20, 20)),
                 _('&About'),
                 _('About pfeMAME'),
                 self.on_about_box_display
                 )
                ]

        # Add each menu bar item from the above list, set any bitmaps, and bind events
        for index, i in enumerate(temp):
            if i[1] != 'separator':
                new_menu_item = wx.MenuItem(i[0], index + 5000, i[2], i[3])
                # Disable the full screen video player menu item if video is disabled
                fvp_id = -1
                if i[4] == self.on_fullscreen_video_player:
                    fvp_id = index + 5000
                if i[1] != '':
                    new_menu_item.SetBitmap(wx.Bitmap(i[1]))
                new_menu_item.SetBackgroundColour(self.application_colors['window_background'])
                new_menu_item.SetTextColour(self.application_colors['text'])
                i[0].Append(new_menu_item)
                if i[4] == self.on_fullscreen_video_player and fvp_id != -1 and self.video_disable:
                    i[0].Enable(fvp_id, False)
                self.Bind(wx.EVT_MENU, i[4], new_menu_item)
            else:
                # Just draw a separator
                i[0].AppendSeparator()

        menu_bar.Append(file_menu, '&File')
        menu_bar.Append(edit_menu, '&Edit')
        menu_bar.Append(view_menu, '&View')
        menu_bar.Append(help_menu, '&Help')
        self.SetMenuBar(menu_bar)
        menu_bar.SetForegroundColour(self.application_colors['text'])
        menu_bar.SetBackgroundColour(self.application_colors['window_background'])
        menu_bar.Bind(wx.EVT_MENU_HIGHLIGHT, on_help_string)
        menu_bar.Bind(wx.EVT_MENU_CLOSE, on_remove_text)
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Menubar'))

        # --------------------------------------------------------------------------------------------------------------
        # Define the sizers for the main window
        self.panel = wx.Panel(self)
        self.panel.SetFont(self.global_font_normal)
        self.panel.SetBackgroundColour(self.application_colors['window_background'])
        # self.panel.BackgroundColour = self.application_colors['window_background']
        self.sizer = wx.BoxSizer(wx.VERTICAL)  # This is the main sizer
        sizer_tb = wx.BoxSizer(wx.HORIZONTAL)  # This is the toolbar sizer
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)  # The 1st nested sizer - its children will be horizontally spaced
        sizer3 = wx.BoxSizer(wx.HORIZONTAL)  # The 2nd nested sizer - its children will be horizontally spaced
        self.sizer4 = wx.BoxSizer(wx.VERTICAL)  # The 3rd nested sizer - its children will be horizontally spaced
        sizer5 = wx.BoxSizer(wx.VERTICAL)  # The 4th nested sizer - its children will be horizontally spaced
        sizer7 = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(sizer_tb, proportion=0, flag=wx.EXPAND)
        self.sizer.Add((10, 2))  # Add some padding to prevent the toolbar from over drawing the next row
        self.sizer.Add(sizer2, proportion=0, flag=wx.EXPAND, border=1)
        self.sizer.Add(sizer3, proportion=1, flag=wx.EXPAND, border=1)
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Main sizers defined'))

        # --------------------------------------------------------------------------------------------------------------
        # Make the message box background colours the same as the main panel
        self.message_window_background_color = self.panel.GetBackgroundColour()

        # --------------------------------------------------------------------------------------------------------------
        # Create key bindings for quick menu access options
        # CTRL-F will trigger a search event from anywhere
        # CTRL-P opens default preferences
        # CTRL-S opens rom specific preferences

        self.Bind(wx.EVT_MENU, self.on_find, id=301)
        self.Bind(wx.EVT_MENU, self.on_fullscreen_video_player, id=304)

        entries = [wx.AcceleratorEntry() for _i in range(4)]
        entries[0].Set(wx.ACCEL_CTRL, ord('F'), 301)
        entries[1].Set(wx.ACCEL_CTRL, ord('P'), 302)
        entries[2].Set(wx.ACCEL_CTRL | wx.ACCEL_SHIFT, ord('P'), 303)
        entries[3].Set(wx.ACCEL_SHIFT, ord('F'), 303)

        accel = wx.AcceleratorTable(entries)
        self.SetAcceleratorTable(accel)

        # --------------------------------------------------------------------------------------------------------------
        # Try to capture the joystick for navigation and determine the directional
        # values for the stick / dpad
        self.stick_ok, self.joystick, self.joystick_polling_timer = functions.joystick_setup(self, self.my_logger, self.pref_joystick_selected)

        # --------------------------------------------------------------------------------------------------------------
        # Define the toolbar
        def toolbar_line_data(primary_image: str, mouseover_image: str, help_text: str, target):
            return (wx.Image(os.path.join(path_to_me, 'images', primary_image), wx.BITMAP_TYPE_PNG).ConvertToBitmap(),
                    wx.Image(os.path.join(path_to_me, 'images', mouseover_image), wx.BITMAP_TYPE_PNG).ConvertToBitmap(),
                    help_text, target, primary_image)

        self.toolbar_data = [
            toolbar_line_data('play.png', 'play_mouseover.png', 'Play ROM', self.on_double_click),
            toolbar_line_data('play_random.png', 'play_random_mouseover.png', 'Play Random ROM', self.on_random),
            toolbar_line_data('filepaths.png', 'filepaths_mouseover.png', 'File Paths', self.on_file_paths),
            toolbar_line_data('verifyrom.png', 'verifyrom_mouseover.png', 'Verify ROM', self.on_verify_roms),
            toolbar_line_data('info.png', 'info_mouseover.png', 'ROM Information', self.on_rom_info),
            toolbar_line_data('tileview.png', 'tileview_mouseover.png', 'Tile View', self.on_tile_view),
            toolbar_line_data('favorite_add.png', 'favorite_add_mouseover.png', 'Add To Favorites', self.on_add_favorite),
            toolbar_line_data('star.png', 'star_mouseover.png', 'ROM Rating', self.on_set_rating),
            toolbar_line_data('refresh.png', 'refresh_mouseover.png', 'Refresh ROMs Database Cache', self.on_database_refresh),
            toolbar_line_data('exit.png', 'exit_mouseover.png', 'Exit', self.on_close)]

        # Change the toolbar button background when you hover over it.  This is really for Linux as Windows does
        # this for you.
        def tool_bar_hover(event) -> None:
            object_selected = event.GetEventObject()
            object_selected.SetBackgroundColour(wx.Colour(229, 241, 251))
            position = functions.match_list_location(self.toolbar_data, 2, object_selected.GetName())
            object_selected.SetBitmap(self.toolbar_data[position][1])  # Replace the main bitmap with the focus bitmap

        # Set the toolbar button background colour back to its default
        def tool_bar_normal(event) -> None:
            object_selected = event.GetEventObject()
            object_selected.SetBackgroundColour(self.application_colors['window_background'])
            position = functions.match_list_location(self.toolbar_data, 2, object_selected.GetName())
            object_selected.SetBitmap(self.toolbar_data[position][0])  # Replace the focus bitmap with the main bitmap

        # Restore the bitmap image back to default when the button is clicked
        def restore_bitmap(event):
            tool_bar_normal(event)

        # Add each toolbar item from the above list, set any bitmaps, and bind events
        for i in self.toolbar_data:
            button = wx.BitmapButton(self.panel, size=(47, 24), bitmap=i[0], style=wx.NO_BORDER, name=i[2])
            button.SetBackgroundColour(self.application_colors['window_background'])
            button.SetBitmapFocus(i[1])
            button.SetToolTip(i[2])
            button.Bind(wx.EVT_ENTER_WINDOW, tool_bar_hover)
            button.Bind(wx.EVT_LEAVE_WINDOW, tool_bar_normal)
            button.Bind(wx.EVT_LEFT_DOWN, restore_bitmap)
            button.Bind(wx.EVT_LEFT_UP, i[3])
            sizer_tb.Add(button, 0, wx.ALL, 2)

        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Toolbar creation'))

        # --------------------------------------------------------------------------------------------------------------
        # Create the 'view' combobox
        text_view = wx.StaticText(self.panel, -1, 'View:', style=wx.ALIGN_CENTRE)
        text_view.SetFont(self.global_font_bold)
        text_view.SetForegroundColour(self.application_colors['text'])
        sizer2.Add(text_view, proportion=0, flag=wx.ALL | wx.ALIGN_BOTTOM, border=3)

        views = [_('All'), _('Favorites'), _('Exists'), _('Missing')]
        self.comboBoxFilter = wx.ComboCtrl(self.panel, wx.ID_ANY, '', size=(200, -1))
        functions.set_dropdown(self.comboBoxFilter, views, 100)
        self.comboBoxFilter.SetBackgroundColour(self.application_colors['window_background'])
        self.comboBoxFilter.SetForegroundColour(self.application_colors['text'])
        self.comboBoxFilter.Bind(wx.EVT_TEXT, self.on_filter_rom_list)  # Value selected binding
        sizer2.Add(self.comboBoxFilter, proportion=0, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=1)

        # --------------------------------------------------------------------------------------------------------------
        # Create a search entry field
        self.clear_event_override = False
        self.controlSearch = SearchCtrl2(self.panel, -1, value='', size=(200, -1), style=0)
        self.controlSearch.SetBackgroundColour(self.application_colors['window_background'])
        self.controlSearch.SetForegroundColour(self.application_colors['text'])
        sizer2.Add(self.controlSearch, proportion=0, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=1)
        self.controlSearch.Bind(wx.EVT_CHAR, self.on_search_dynamic)

        # --------------------------------------------------------------------------------------------------------------
        # Create the versions text fields
        text = wx.StaticText(self.panel, -1, _('MAME Version: '), style=wx.ALIGN_LEFT)
        text.SetFont(self.global_font_bold)
        text.SetForegroundColour(self.application_colors['text'])
        text_version_mame = wx.StaticText(self.panel, -1, _('Unknown'), style=wx.ALIGN_LEFT)
        text_version_mame.SetForegroundColour(self.application_colors['text'])
        sizer2.Add(width=1, height=1, proportion=1, flag=wx.EXPAND)
        sizer2.Add(text, proportion=0, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=1)
        sizer2.Add(text_version_mame, proportion=0, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=1)

        # --------------------------------------------------------------------------------------------------------------
        # Create the database age text fields
        text_database_age = wx.StaticText(self.panel, -1, _('Database Age: '), style=wx.ALIGN_LEFT)
        text_database_age.SetFont(self.global_font_bold)
        text_database_age.SetForegroundColour(self.application_colors['text'])
        self.text_database_ageValue = wx.StaticText(self.panel, -1, _('Unknown'), style=wx.ALIGN_LEFT)
        self.text_database_ageValue.SetForegroundColour(self.application_colors['text'])
        sizer2.Add(text_database_age, proportion=0, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=1)
        sizer2.Add(self.text_database_ageValue, proportion=0, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=1)

        # --------------------------------------------------------------------------------------------------------------
        # Show the image option selector
        self.image_size_x = int(224 * 1.2)
        self.image_size_y = int(256 * 1.2)
        image_option_sizer = wx.BoxSizer(wx.HORIZONTAL)
        btn_size = (20, 22)
        tmp = wx.ArtProvider.GetBitmap(wx.ART_GO_BACK, wx.ART_BUTTON, btn_size)
        btn_left = wx.StaticBitmap(self.panel, -1, tmp, size=btn_size)
        btn_left.Bind(wx.EVT_LEFT_DOWN, self.image_options_left)

        self.image_option = wx.StaticText(self.panel, -1, _('Empty'), size=(self.image_size_x - 40, 20), style=wx.TE_CENTER)
        self.image_option.SetForegroundColour(self.application_colors['text'])
        if self.image_options:  # Don't do this if the data is not initialised
            self.image_option.SetLabel(self.image_options[self.image_options_index][1])

        tmp = wx.ArtProvider.GetBitmap(wx.ART_GO_FORWARD, wx.ART_BUTTON, btn_size)
        btn_right = wx.StaticBitmap(self.panel, -1, tmp, size=btn_size)
        btn_right.Bind(wx.EVT_LEFT_DOWN, self.image_options_right)

        image_option_sizer.Add(btn_left)
        image_option_sizer.Add(self.image_option, flag=wx.ALIGN_CENTER_VERTICAL)
        image_option_sizer.Add(btn_right)
        sizer7.Add(image_option_sizer, proportion=0, flag=wx.EXPAND, border=0)

        # --------------------------------------------------------------------------------------------------------------
        # Create the ROM list control
        self.list_notebook = flatnotebook.FlatNotebook(self.panel, agwStyle=flatnotebook.FNB_NO_NAV_BUTTONS |
                                                                  flatnotebook.FNB_NO_X_BUTTON |
                                                                  flatnotebook.FNB_TABS_BORDER_SIMPLE |
                                                                  flatnotebook.FNB_NODRAG)
        self.list_notebook.Bind(flatnotebook.EVT_FLATNOTEBOOK_PAGE_CHANGED, self.notebook_page_changed)
        self.list_notebook.SetActiveTabColour(self.application_colors['window_background'])
        self.list_notebook.SetTabAreaColour(self.application_colors['window_background'])
        self.list_notebook.SetNonActiveTabTextColour(self.application_colors['text'])
        self.list_notebook.SetForegroundColour(self.application_colors['text'])
        self.list_notebook.SetBackgroundColour(self.application_colors['window_background'])
        self.list_notebook.SetActiveTabTextColour(self.application_colors['text'])
        self.list_notebook_current = 0  # Define that the first tab is currently shown

        self.list = VirtualList(self.list_notebook,
                                self.pref_icon_size,
                                self.pref_show_rom_icons,
                                self.folders_icons,
                                self.pref_highlight_missing,
                                self.application_colors,
                                self.my_logger)  # This is the virtual list control class

        self.list.SetFont(self.global_font_normal_8)
        self.list.Bind(ulc.EVT_LIST_ITEM_SELECTED, self.on_single_click)  # Create the single click event binding
        self.list.Bind(ulc.EVT_LIST_ITEM_ACTIVATED, self.on_double_click)  # Create the double click event binding
        self.list.Bind(wx.EVT_RIGHT_DOWN, self.on_right_down)  # Bind the right click to the ROMs list control
        self.list.Bind(ulc.EVT_LIST_ITEM_FOCUSED, self.on_page_changed)  # Make sure the list redraws properly
        self.list.Bind(wx.EVT_CHAR_HOOK, self.on_key_press)  # Bind a keypress event
        self.list.Bind(wx.EVT_LIST_COL_CLICK, self.on_column_clicked)
        self.list.Bind(wx.EVT_MOTION, self.on_mouse)

        self.list_notebook.AddPage(self.list, _('ROMs List'))
        self.list_notebook.AddPage(preferences.Preferences(parent=self.list_notebook,
                                                           title=_('Default Preferences'),
                                                           mode='default',
                                                           rom_file_name='',
                                                           folder_ini='',
                                                           path_executable='',
                                                           system_file_name=''), _('pfeMAME Preferences'))

        self.sizer4.Add(self.list_notebook, proportion=1, flag=wx.EXPAND, border=1)  # Add the control to the sizer

        # --------------------------------------------------------------------------------------------------------------
        # Capture the window resize event
        self.Bind(wx.EVT_SIZE, self.on_resize_window)

        # --------------------------------------------------------------------------------------------------------------
        # Insert a placeholder for the ROM image - this is just a blank bitmap for now
        self.image_panel = ImagePanel(self.panel, self.image_size_x, self.image_size_y, self.list)
        self.image_panel.Bind(wx.EVT_MOTION, self.on_mouse)
        sizer7.Add(self.image_panel, proportion=0, border=0)

        # --------------------------------------------------------------------------------------------------------------
        # Add the media player control
        # It starts off hidden as it is designed to occupy the same space as the ROM image (one will hide and the other
        # will show)
        if not self.video_disable:
            if self.use_vlc:
                # This is the code to set up the VLC backend for video playback
                self.my_logger.info(_('Trying to use VLC as the video playback device'))  # Write to the log file
                try:
                    self.pnlVideo = wx.Panel(self.panel, wx.ID_ANY)
                    self.instance = vlc.Instance('--quiet --autoscale --fullscreen')
                    self.player = self.instance.media_player_new()
                    if self.system == 'windows':
                        self.player.set_hwnd(self.pnlVideo.GetHandle())
                    elif self.system == 'mac':
                        self.player.set_nsobject(self.pnlVideo.GetHandle())
                    else:
                        self.player.set_xwindow(self.pnlVideo.GetHandle())

                    # Bind the event for VLC video ended so that we can flag to exit the wait loop
                    # and show the ROM image
                    mp_em = self.player.event_manager()
                    mp_em.event_attach(vlc.EventType.MediaPlayerEndReached, self.on_media_stopped_vlc)

                    sizer7.Add(self.pnlVideo, proportion=0, flag=wx.ALL, border=1)
                    self.pnlVideo.Hide()  # Start with it hidden as it's going to sit over the top of the ROM image
                    self.pnlVideo.Bind(wx.EVT_MOTION, self.on_mouse)
                    self.pnlVideo.SetSize(self.image_size_x, self.image_size_y)
                except NameError:
                    # It is likely that VLC doesn't exist on this platform
                    print(_('It is likely that VLC doesnt exist on this platform.  Falling back to wx MediaCtrl'))
                    self.my_logger.info(_('It is likely that VLC doesnt exist on this platform.  Falling back to wx MediaCtrl'))
                    self.use_vlc = False

            else:
                try:
                    self.my_logger.info(
                        _('Trying to use wx MediaCtrl as the video playback device'))  # Write to the log file
                    # This is the code to set up the wx MediaCtrl for video playback.
                    self.mc = functions.mc_video_player_setup(self,
                                                              self.system,
                                                              self.panel,
                                                              self.image_size_x,
                                                              self.image_size_y,
                                                              self.on_media_loaded,
                                                              self.on_media_stopped)
                    sizer7.Add(self.mc, proportion=0, flag=wx.ALL, border=1)
                    self.mc.Hide()  # Start with it hidden as it's going to sit over the top of the ROM image control
                    self.mc.Bind(wx.EVT_MOTION, self.on_mouse)
                except NotImplementedError:
                    self.my_logger.info(_('No media content player available on this system'))  # Write to the log file
                    self.media_player_available = False
        
            self.videoTimer = wx.Timer(self)  # Used to autoplay ROM videos
            self.timerSeconds = 5  # Always start with a 5-second timer
            self.Bind(wx.EVT_TIMER, self.on_timer_event, self.videoTimer)
        else:
            self.my_logger.info(_('Video playback disable selected in preferences or via command line'))  # Write to the log file
            self.media_player_available = False

        self.videoLoadCaption = wx.Button(self.panel, -1, label=_('No Video Available'), size=(224, 20))
        self.videoLoadCaption.Bind(wx.EVT_BUTTON, self.on_play_clicked)
        
        sizer7.Add(self.videoLoadCaption, proportion=0, flag=wx.EXPAND, border=0)

        # --------------------------------------------------------------------------------------------------------------
        # Add the ROM info and MAME info text controls.  These are placed inside a notebook.
        notebook = flatnotebook.FlatNotebook(self.panel, agwStyle=flatnotebook.FNB_NO_NAV_BUTTONS |
                                                                  flatnotebook.FNB_NO_X_BUTTON |
                                                                  flatnotebook.FNB_TABS_BORDER_SIMPLE |
                                                                  flatnotebook.FNB_NODRAG)

        notebook.SetActiveTabColour(self.application_colors['window_background'])
        notebook.SetTabAreaColour(self.application_colors['window_background'])
        notebook.SetNonActiveTabTextColour(self.application_colors['text'])
        notebook.SetForegroundColour(self.application_colors['text'])
        notebook.SetActiveTabTextColour(self.application_colors['text'])

        self.controlRomInfo = wx.TextCtrl(notebook, 
                                          value='', 
                                          size=(224, 310),
                                          style=wx.TE_MULTILINE
                                              | wx.TE_READONLY
                                              | wx.TE_AUTO_URL
                                              | wx.TE_RICH
                                              | wx.NO_BORDER
                                              | wx.EXPAND)
        self.controlRomInfo.Bind(wx.EVT_TEXT_URL, self.on_url_selected)
        self.controlRomInfo.Bind(wx.EVT_MOTION, self.on_mouse)
        self.controlRomInfo.SetFont(self.global_font_normal_8)
        self.controlRomInfo.SetBackgroundColour(self.message_window_background_color)
        self.controlRomInfo.SetForegroundColour(self.application_colors['text'])

        self.controlMAMEInfo = wx.TextCtrl(notebook, 
                                           value='', 
                                           size=(224, 310),
                                           style=wx.TE_MULTILINE
                                               | wx.TE_READONLY
                                               | wx.TE_AUTO_URL
                                               | wx.TE_RICH
                                               | wx.NO_BORDER
                                               | wx.EXPAND)
        self.controlMAMEInfo.Bind(wx.EVT_MOTION, self.on_mouse)
        self.controlMAMEInfo.SetFont(self.global_font_normal_8)
        self.controlMAMEInfo.SetBackgroundColour(self.message_window_background_color)
        self.controlMAMEInfo.SetForegroundColour(self.application_colors['text'])

        notebook.AddPage(self.controlRomInfo, _('ROM Info'))
        notebook.AddPage(self.controlMAMEInfo, _('MAME Info'))

        sizer7.Add(notebook, proportion=0, flag=wx.EXPAND, border=0)

        # --------------------------------------------------------------------------------------------------------------
        # Add the messages box
        text2 = wx.StaticText(self.panel, -1, _('Messages:'))
        text2.SetFont(self.global_font_bold)
        text2.SetForegroundColour(self.application_colors['text'])
        text2.SetBackgroundColour(self.application_colors['text_background'])
        self.controlMessages = wx.TextCtrl(self.panel, 
                                           value='', 
                                           size=(800, 140),
                                           style=wx.TE_MULTILINE 
                                               | wx.TE_READONLY
                                               | wx.NO_BORDER
                                               | wx.EXPAND)
        self.controlMessages.Bind(wx.EVT_LEFT_DCLICK, self.disable_alert_timer)
        self.controlMessages.Bind(wx.EVT_MOTION, self.on_mouse)
        self.controlMessages.SetFont(self.global_font_normal_8)
        self.controlMessages.SetBackgroundColour(self.message_window_background_color)
        self.controlMessages.SetForegroundColour(self.application_colors['text'])

        sizer5.Add(text2, proportion=0, flag=wx.EXPAND, border=1)
        sizer5.Add(self.controlMessages, proportion=1, flag=wx.EXPAND, border=1)

        self.sizer4.Add(sizer5, proportion=0, flag=wx.EXPAND, border=1)
        sizer3.Add(self.sizer4, proportion=1, flag=wx.EXPAND, border=1)

        # --------------------------------------------------------------------------------------------------------------
        # Add the ROM categories box
        self.panel1 = wx.lib.scrolledpanel.ScrolledPanel(self.panel, -1, size=(150, -1))
        self.panel1.SetupScrolling(scroll_x=False)  # Suppress the horizontal scroll bar
        self.panel1.Bind(wx.EVT_LEFT_DCLICK, self.on_doubleclick_categories)
        self.panel1.Bind(wx.EVT_MOTION, self.on_mouse)
        self.panel1.SetFont(self.global_font_normal_8)
        self.panel1.SetForegroundColour(self.application_colors['text'])
        self.panel1.SetBackgroundColour(self.application_colors['text_background'])

        text2 = wx.StaticText(self.panel, -1, _('ROM Categories:'))
        text2.SetFont(self.global_font_bold)
        text2.SetForegroundColour(self.application_colors['text'])
        text2.SetBackgroundColour(self.application_colors['text_background'])
        svgt = wx.BoxSizer(wx.VERTICAL)
        sizer3.Add(svgt, proportion=0, flag=wx.EXPAND, border=1)
        svgt.Add(text2, proportion=0, flag=wx.EXPAND, border=1)
        svgt.Add(self.panel1, 1, wx.EXPAND | wx.ALL)
        sizer_vertical_romtypes = wx.BoxSizer(wx.VERTICAL)
        self.panel1.SetSizer(sizer_vertical_romtypes)
        sizer3.Add(sizer7, proportion=0, flag=wx.EXPAND, border=1)
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Controls creation'))

        # --------------------------------------------------------------------------------------------------------------
        # Add the override for wx.Log messages
        wx.Log.SetActiveTarget(functions.MyLog(self.controlMessages, self.my_logger))  # All wx.Log messages will go to the messages box

        # --------------------------------------------------------------------------------------------------------------
        # Add the custom status bar panel
        self.statusbar_panel = functions.StatusBarPanel(self.panel)
        self.sizer.Add(self.statusbar_panel, proportion=0, flag=wx.EXPAND)

        # --------------------------------------------------------------------------------------------------------------
        # Get the window to size automatically based upon sizer requirements
        self.SetMinSize(self.sizer.GetMinSize())  # Prevent you from making smaller than the default size
        self.panel.SetSizer(self.sizer)
        self.sizer.Fit(self)

        # --------------------------------------------------------------------------------------------------------------
        # Initialise the progress bar
        self.progress_bar = gauge.PyGauge(self.panel, -1, style=wx.GA_HORIZONTAL | wx.GA_SMOOTH | wx.GA_PROGRESS)
        self.progress_bar.SetBarColor([wx.Colour(162, 255, 178), wx.Colour(159, 176, 255)])

        posx, posy = self.statusbar_panel.GetPosition()
        sizex, sizey = self.statusbar_panel.GetSize()
        self.progress_bar.SetPosition((posx + 2, posy + 2))
        self.progress_bar.SetSize((sizex - 4, sizey - 4))
        self.progress_bar.Hide()  # Hide the progress bar until it is needed

        # --------------------------------------------------------------------------------------------------------------
        # Position the frame on screen and make it visible
        self.SetPosition(wx.Point(0, 0))
        self.Show(True)

        self.Update()  # Make sure everything is visible just in case an auto refresh occurs
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Set size and display window'))

        # --------------------------------------------------------------------------------------------------------------
        # Pre-load the history file (Assuming it exists)
        self.load_history_dat()
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'History file loaded'))

        # --------------------------------------------------------------------------------------------------------------
        # Pre-load the mameinfo.dat file (Assuming it exists)
        self.load_mameinfo_dat()
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'mameinfo.dat file loaded'))

        # --------------------------------------------------------------------------------------------------------------
        # Pre-load the categories file (Assuming it exists)
        self.load_categories()
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Categories created'))

        # --------------------------------------------------------------------------------------------------------------
        # If we have a MAME executable path defined, then try to execute the '-?' command to get the version info
        self.mame_version = _('Unknown')
        if self.path_mame_exe != '':
            try:
                # Build up the command
                cmd = [self.path_mame_exe, '-?']
                p = subprocess.run(cmd, stdout=subprocess.PIPE, shell=self.shell_to_use, universal_newlines=True)
                out = p.stdout
                self.mame_version = out[out.find('v') + 1: out.find('(') - 1]
                if self.mame_version == '':
                    self.mame_version = _('Unknown')
            except FileNotFoundError:
                self.mame_version = _('Unknown')
            text_version_mame.SetLabel(self.mame_version)
            self.my_logger.info(f'MAME version: {self.mame_version}')  # Write to the log file
            self.my_logger.info(f'pfeMAME version: {__version__}')  # Write to the log file
            self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                            _('Retrieved MAME version')))

        # --------------------------------------------------------------------------------------------------------------
        # Populate the game categories window
        # Look at the categories.dat file to see what (if any) game filter categories have been set
        # If they are set, then check the corresponding category checkbox when it is drawn
        # Then we create the checkbox either set or cleared and add it to the sizer
        self.category_dict, self.catver_version = self.build_categories()
        self.Freeze()

        for line in sorted(set(self.category_dict.values())):
            cb = wx.CheckBox(self.panel1, label=line, name=line)
            cb.Bind(wx.EVT_CHECKBOX, self.on_update_categories)  # Create the single click event binding
            cb.Bind(wx.EVT_MOTION, self.on_mouse)
            cb.SetValue(line in self.categories)
            sizer_vertical_romtypes.Add(cb, 0, wx.ALL)  # Add the control to the sizer
        sizer_vertical_romtypes.Add((10, 10))
        self.Thaw()
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'ROM categories window populated'))

        # --------------------------------------------------------------------------------------------------------------
        # Check to see if there are any software updates available
        if self.pref_check_software_version == 'True':
            self.check_for_updates('auto')
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Starting software update check process'))

        # --------------------------------------------------------------------------------------------------------------
        # Check if the application restarted from an internal forced restart
        if forced_restart != '':
            self.my_logger.info(forced_restart)
            self.messages_append(f'{forced_restart}\n', False)

        # --------------------------------------------------------------------------------------------------------------
        # All startup actions complete
        msg = _('Application startup complete in {:.2f} seconds').format((time.time() - self.startup_process_start_time))
        self.my_logger.info(msg)
        self.messages_append(f'{msg}\n', True)
        self.start_up_complete = True  # Used by the single click event handler

        # --------------------------------------------------------------------------------------------------------------
        # Call the processes that must happen after the main display is drawn.  These include database setup, checks etc 
        wx.CallAfter(self.initial_setup)

        # --------------------------------------------------------------------------------------------------------------
        # Give the ROMs list the focus
        self.list.SetFocus()
        self.Refresh()

        # --------------------------------------------------------------------------------------------------------------
        # If we are in developer mode then display developer startup information
        if self.development is True:
            self.messages_append(_('Developer mode enabled\n'))
            for line in self.development_startup_info:
                self.messages_append(line)

        # --------------------------------------------------------------------------------------------------------------
        # Capture mouse events just so that we can reset the inactivity timer if the mouse is moved
        self.panel.Bind(wx.EVT_MOTION, self.on_mouse)
        self.panel.Bind(wx.EVT_LEFT_DOWN, self.on_mouse)
        self.panel.Bind(wx.EVT_RIGHT_DOWN, self.on_mouse)

    # --------------------------------------------------------------------------------------------------------------
    # Process initial setups.  These are called after the main display is drawn as some of them call long-running
    # tasks that would otherwise cause the display to lockup and not look right.
    def initial_setup(self) -> None:
        # --------------------------------------------------------------------------------------------------------------
        # Build up the ROM database, parse various flags, and format for final display
        # wx.CallLater(self.on_database_refresh, 'auto')
        self.on_database_refresh('auto')
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'ROMs database loaded'))

        # --------------------------------------------------------------------------------------------------------------
        # If the preferences setting for recalling the last selected ROM is set, then do it
        if self.lastviewselected != '' \
                and self.prefUseLastSelected == 'True' \
                and self.lastromselected != '':
            # Recall the previous 'view' setting
            self.comboBoxFilter.SetValue(self.lastviewselected)
            self.on_filter_rom_list(-1)  # Force an update of the ROM list
            
            # Local copies of variables to make the for loops below faster
            last_rom_selected = self.lastromselected
            last_system_selected = self.lastsystemselected

            # Look for a match with the 'lastromselected' from the INI file.  If found, set this in the current view.
            match_index = next((i for i, x in enumerate(MainFrame.mylist_filtered) if
                                x[index_rom_name] == last_rom_selected), None)

            # If we found a match above, now check to see if it was a system program.  If it was, look for a match
            # with the system program name - ensure you look after the current match point to prevent a match with
            # a program name from another system
            if self.lastsystemselected != '' and match_index is not None:
                start_index = match_index  # Start the search after the main system program name
                match_index = next((i for i, x in enumerate(MainFrame.mylist_filtered[start_index + 1:]) if
                                    x[index_system_program_file_name] == last_system_selected),
                                   None)
                if match_index is not None:
                    match_index += start_index + 1  # Combine the search result indexes to get the right index

            if match_index is not None:
                self.list.Select(int(match_index))
                self.list.EnsureVisible(int(match_index))
        else:
            # ----------------------------------------------------------------------------------------------------------
            # Set the filter view based upon preferences and force a refresh
            self.comboBoxFilter.SetValue(self.prefMainView)  # Set as per the preferences
            self.on_filter_rom_list(-1)  # Force an update of the ROM list
            # There won't be any ROM selected, so we need to just select the first ROM in the list
            if len(self.mylist) > 0:
                self.list.Select(0)
                self.list.EnsureVisible(0)
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Last rom selected process'))

        # --------------------------------------------------------------------------------------------------------------
        # Adjust the window size and show / hide the systems column depending upon the preferences setting
        if self.prefSystemsSupport == 'False':
            self.list.SetColumnShown(index_system_program_file_name, False)
            self.list.SetColumnShown(index_system_short_name, False)
            # self.SetSize(self.GetSize() - (228, 0))

        # --------------------------------------------------------------------------------------------------------------
        # Hide this column as we never want to see it
        self.list.SetColumnShown(index_path_media_name, False)

        # --------------------------------------------------------------------------------------------------------------
        # Check to see if our preferences setting for the tile view is set.
        # If it is, then after startup we need to automatically switch to the tile view
        if self.prefView == 'True':
            self.inactivity_timer.Stop()  # Stop the inactivity timer from this view
            self.on_tile_view(-1)  # Launch tile view

        # --------------------------------------------------------------------------------------------------------------
        # Check when the ROMs database cache was fully updated - suggest a refresh if its getting old
        # If it's blank then skip this as it will be fully refreshed
        if self.cacheDate != '' and (self.prefCheckDatabaseAge == 'True' or self.prefCheckDatabaseAge == ''):
            try:
                time1 = datetime.datetime.strptime(self.cacheDate, '%Y-%m-%d %H:%M:%S')
                today = datetime.datetime.strptime(self.today, '%Y-%m-%d %H:%M:%S')
                days_old = (today - time1).days
                if days_old > int(self.prefDaysCheckCache):
                    dlg = wx.RichMessageDialog(self,
                                               _('The database cache is {} days old.  '
                                                 'I suggest you refresh the ROMs database cache manually.').format(str(days_old)),
                                               _('ROMs Database Cache out of date'),
                                               wx.OK | wx.ICON_INFORMATION)
                    dlg.ShowCheckBox(_('Dont show this message again'))
                    dlg.ShowModal()
                    if dlg.IsCheckBoxChecked():
                        # Update the database date in the INI file
                        iniFile.set_ini_file_variables([('preferences', 'checkdatabaseage', 'False')])
                    dlg.Destroy()
                    self.Refresh()
                # Update the ROMs database cache age display
                self.text_database_ageValue.SetLabel(f'{str(days_old)} days')
            except ValueError:
                dlg = wx.MessageDialog(self,
                                       _('Unable to check age of ROMs cache.  '
                                       'Please check the Preferences and then refresh the ROMs cache'),
                                       _('ROMs Database Cache check error'),
                                       wx.OK | wx.ICON_INFORMATION)
                dlg.ShowModal()
                dlg.Destroy()
                self.Refresh()
        self.development_startup_info.append(functions.debug_type_print(self.development, self.debug_start_time,
                                                                        'Database cache age check'))

        # --------------------------------------------------------------------------------------------------------------
        # Compare the catver.ini file version against the MAME version and advise the user if there is a mismatch
        if self.category_dict != {} and self.pref_check_catver_version == 'True' and self.path_mame_exe != '':
            if functions.compare_version(self.mame_version, self.catver_version) == 1:
                dlg = wx.RichMessageDialog(self,
                                           _('Your MAME version is {} while your catver.ini file'
                                             ' version is {}.  Suggest you find an updated copy'
                                             ' of the catver.ini file if it is older than the MAME version.').format(self.mame_version, self.catver_version),
                                           _('MAME and catver version mismatch'),
                                           wx.OK | wx.ICON_INFORMATION)
                dlg.ShowCheckBox(_('Don\'t show this message again'))
                dlg.ShowModal()
                if dlg.IsCheckBoxChecked():
                    # Update the database date in the INI file
                    iniFile.set_ini_file_variables([('preferences', 'checkcatverversion', 'False')])
                    pub.sendMessage('disable_catver_check')
                dlg.Destroy()
                self.Refresh()

        # --------------------------------------------------------------------------------------------------------------
        # If we have received the command line option to manually run a specific rom on startup, then do it now
        if self.run_specified_rom is True:
            # Set the position in the ROMs list based on the rom name specified
            match_index = next((i for i, x in enumerate(MainFrame.mylist_filtered) if
                                x[index_rom_file_name] == self.specified_rom_name), None)
            if match_index is not None:
                self.list.Select(int(match_index))
                self.list.EnsureVisible(int(match_index))
            # Run the rom
            self.on_double_click(suppress=True,
                                 ignore_columns=False,
                                 input_playback_or_record='',
                                 pb_r_filename='',
                                 rom_file_name=self.specified_rom_name,
                                 system_file_name='')

    # ------------------------------------------------------------------------------------------------------------------
    # Call this event when we have selected a different tab on the list notebook
    def notebook_page_changed(self, _event) -> None:
        # Check to see if we have three notebook pages.  If we do, the 3rd page is the ROM specific.  If we don't have
        # that pages selected, then delete it.
        if self.list_notebook.GetPageCount() > 2:
            if self.list_notebook.GetSelection() != 2:
                self.list_notebook.DeletePage(2)

    # ------------------------------------------------------------------------------------------------------------------
    # Reset the inactivity timer on any mouse event
    def on_mouse(self, _event) -> None:
        self.inactivity_timer_start()  # Reset the inactivity timer

    # ------------------------------------------------------------------------------------------------------------------
    # Process the joystick polling timer so we can see if the POV HAT / DPAD direction has been selected.
    def on_joystick_pov(self, _event) -> None:
        if not self.tileview_running and not self.joystick_block and self.pref_enable_joystick_mainview == 'True':
            pov = self.joystick.GetPOVPosition()
            if pov != 65535:
                # POV direction detected
                self.inactivity_timer_start()  # Reset the inactivity timer

                if pov == 0:  # Have we selected move up?
                    self.scroll_list_up()  # scroll up a row
                if pov == 18000:  # Have we selected move down?
                    self.scroll_list_down()  # scroll down a row

    # ------------------------------------------------------------------------------------------------------------------
    # This function processes joystick events for navigation.
    # It will determine which joystick button / direction has been selected
    # and will call the required function essentially mapping the joystick
    # to the keyboard arrow keys and buttons
    def on_joystick(self, _event) -> None:
        self.inactivity_timer_start()  # Reset the inactivity timer
        # Don't allow these events to process if we are running tileview in the foreground
        if not self.tileview_running and not self.joystick_block and self.pref_enable_joystick_mainview == 'True':
            # Get the button and direction raw values from the joystick
            button_value = self.joystick.GetButtonState()
            direction_value = self.joystick.GetPosition()

            # Process the buttons
            if button_value == 1:
                # Joystick button for 'play game' pressed
                self.on_double_click(None)  # play the selected game
            elif button_value == 2:
                # Joystick 2nd button - do nothing for now
                pass
            elif button_value == 32:
                # Right bumper selected so page down the list
                self.scroll_list_down_page()
            elif button_value == 16:
                # Left bumper selected so page up the list
                self.scroll_list_up_page()

            # Process the directions
            # This convoluted method is due to the different way that windows & linux determine joystick
            # positions.  E.g. Linux is -327676 - 0 - 32767 for the left, center, and right
            # while Windows is 0 - 32767 - 65535 for the left, center, and right
            # x_position = direction_value[0]
            y_position = direction_value[1]
            # x_center = ((self.joystickXMax - self.joystickXMin) // 2)//2
            y_center = ((self.joystickYMax - self.joystickYMin) // 2) // 2

            if y_position < (self.joystickYMin + y_center):  # Have we selected move up?
                self.scroll_list_up()  # scroll up a row
            if y_position > (self.joystickYMax - y_center):  # Have we selected move down?
                self.scroll_list_down()  # scroll down a row

    # ------------------------------------------------------------------------------------------------------------------
    # Reset the pfeMAME INI file back to install default values. 
    # This is just a fallback option if things are not working right and you want to start again with your settings.
    def on_reset_pfemame_ini(self, _event) -> None:
        dlg = wx.MessageDialog(self,
                               _('Reset the pfeMAME INI file to its default settings?  '
                               'WARNING:  This resets any changes you have made!'),
                               _('Reset pfeMAME INI file'),
                               style=wx.YES_NO | wx.NO_DEFAULT | wx.ICON_QUESTION)
        if dlg.ShowModal() == wx.ID_YES:
            iniFile.create_ini_file()
            dlg2 = wx.MessageDialog(self,
                                    _('Please restart pfeMAME for the new default INI file settings to be used.'),
                                    _('Restart pfeMAME'),
                                    style=wx.OK)
            dlg2.ShowModal()
            dlg2.Destroy()
        dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # Enable the alert color strobing timer
    def enable_alert_timer(self) -> None:
        if self.pref_flash_messages_if_error == 'True:':
            self.alert_timer.Start(50)
            msg = _('IMPORTANT NOTIFICATION:  '
                   'Please check this window for the ERROR message.  '
                   'Double-click to cancel alert.')
            self.messages_append(f'{msg}\n', True)

    # ------------------------------------------------------------------------------------------------------------------
    # Disable the alert color strobing timer
    def disable_alert_timer(self, _event) -> None:
        self.alert_timer.Stop()
        self.controlMessages.SetBackgroundColour(self.message_window_background_color)
        self.controlMessages.Refresh()

    # ------------------------------------------------------------------------------------------------------------------
    # Get the mame.ini and pfemame.ini file variables
    def setup_ini_variables(self, msg: str = '') -> None:
        self.my_logger.info(_('Reading INI file settings'))  # Write to the log file

        # Check if the INI file actually exists
        if not os.path.isfile(os.path.join(path_to_me, 'pfemame.ini')):
            # For some reason the pfemame.ini file doesn't exist, so we need to create it
            self.my_logger.error(_('Unable to find pfeMAME.ini file'))  # Write to the log file
            iniFile.create_ini_file()
            self.my_logger.info(_('New pfeMAME.ini file created'))  # Write to the log file

        # Read variables from the pfemame.ini file
        self.cacheDate = iniFile.read_ini_file('system', 'cachedate')
        self.lastromselected = iniFile.read_ini_file('system', 'lastromselected')
        self.lastsystemselected = iniFile.read_ini_file('system', 'lastsystemselected')
        self.lastviewselected = iniFile.read_ini_file('system', 'lastviewselected')
        self.prefCheckDatabaseAge = iniFile.read_ini_file('preferences', 'checkdatabaseage')
        self.prefUseLastSelected = iniFile.read_ini_file('preferences', 'uselastselected')
        self.pref_icon_size = iniFile.read_ini_file('preferences', 'iconsize', default='16')
        self.pref_show_rom_icons = iniFile.read_ini_file('preferences', 'showgameicons')
        self.prefMainView = iniFile.read_ini_file('preferences', 'mainview', default='All')
        self.pref_highlight_missing = iniFile.read_ini_file('preferences', 'highlitemissing')
        self.prefDaysCheckCache = iniFile.read_ini_file('preferences', 'dayscheckcache', default='5')
        self.prefScreenSaverTime = int(iniFile.read_ini_file('preferences', 'screensavertime', default='10'))
        self.folders_icons = iniFile.read_ini_file('folders', 'icons')
        self.foldersVideo = iniFile.read_ini_file('folders', 'video')
        self.folder_mapping = iniFile.read_ini_file('folders', 'input_mapping')
        self.path_mame_exe = iniFile.read_ini_file('path', 'mame_executable')
        self.pathHistory = iniFile.read_ini_file('path', 'history_dat')
        self.pathMAMEInfoDat = iniFile.read_ini_file('path', 'mameinfo_dat')
        self.pathCatver = iniFile.read_ini_file('path', 'catver_ini', default=os.path.join(path_to_me, 'catver.ini'))
        self.prefView = iniFile.read_ini_file('preferences', 'view')
        self.prefSystemsSupport = iniFile.read_ini_file('preferences', 'systemssupport')
        self.prefVideoAutoload = iniFile.read_ini_file('preferences', 'videoautoload')
        self.prefVideoDisable = iniFile.read_ini_file('preferences', 'videodisable')
        # If the Video Disable is True then also set the global self.video_disable true as that is used everywhere
        if self.prefVideoDisable == 'True':
            self.video_disable = True
        self.prefClickFavRating = iniFile.read_ini_file('preferences', 'clickfavrating')
        self.pref_short_name_global_suppress = iniFile.read_ini_file('preferences', 'suppressshortname')
        self.pref_use_multiprocessor = iniFile.read_ini_file('preferences', 'usemultiprocessor')
        self.pref_full_systems_method = iniFile.read_ini_file('preferences', 'fullsystemsmethod')
        self.pref_full_systems_method = True if self.pref_full_systems_method == 'True' else False
        self.path_to_mame_ini_file = iniFile.read_ini_file('path', 'mame_ini')
        self.application_colors = functions.get_application_colors()
        self.pref_check_catver_version = iniFile.read_ini_file('preferences', 'checkcatverversion')
        self.pref_check_software_version = iniFile.read_ini_file('preferences', 'checksoftwareversion', default='True')
        self.pref_flash_messages_if_error = iniFile.read_ini_file('preferences', 'flashiferrors', default='True')
        self.pref_joystick_selected = iniFile.read_ini_file('preferences', 'joystickselected')
        self.pref_enable_joystick_mainview = iniFile.read_ini_file('preferences', 'enablejoystickmainview', default='True')

        fonts = functions.get_font_info()
        self.global_font_bold = fonts.global_font_bold
        self.global_font_normal = fonts.global_font_normal
        self.global_font_normal_8 = fonts.global_font_normal_8
        self.global_font_normal_12 = fonts.global_font_normal_12

        # Read variables from the mame.ini file
        self.folders_roms = iniFile.get_from_mameini('rompath', self.path_to_mame_ini_file)
        self.folders_systems = self.folders_roms
        self.folders_samples = iniFile.get_from_mameini('samplepath', self.path_to_mame_ini_file)
        self.foldersSnap = iniFile.get_from_mameini('snapshot_directory', self.path_to_mame_ini_file)
        self.folders_hash = iniFile.get_from_mameini('hashpath', self.path_to_mame_ini_file)
        self.folder_ini = iniFile.get_from_mameini('inipath', self.path_to_mame_ini_file)
        self.folderInput = iniFile.get_from_mameini('input_directory', self.path_to_mame_ini_file)
        self.folder_config = iniFile.get_from_mameini('cfg_directory', self.path_to_mame_ini_file)

        # Try to read variables from the ui.ini file and create the image options list
        self.image_options = []
        if self.folder_ini != '' and os.path.isdir(self.folder_ini):
            self.pcbs_directory = iniFile.get_from_mameini('pcbs_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.bosses_directory = iniFile.get_from_mameini('bosses_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.cabinets_directory = iniFile.get_from_mameini('cabinets_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.cpanels_directory = iniFile.get_from_mameini('cpanels_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.flyers_directory = iniFile.get_from_mameini('flyers_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.titles_directory = iniFile.get_from_mameini('titles_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.ends_directory = iniFile.get_from_mameini('ends_directory', os.path.join(self.folder_ini, 'ui.ini'))
            self.marquees_directory = iniFile.get_from_mameini('marquees_directory', os.path.join(self.folder_ini, 'ui.ini'))
            dir_list = [[self.foldersSnap, 'Snapshots'],
                        [self.pcbs_directory, 'PCBs'],
                        [self.bosses_directory, 'Bosses'],
                        [self.cabinets_directory, 'Cabinets'],
                        [self.cpanels_directory, 'Control Panels'],
                        [self.flyers_directory, 'Flyers'],
                        [self.titles_directory, 'Title Screens'],
                        [self.ends_directory, 'Ending Screens'],
                        [self.marquees_directory, 'Marquees']]
        else:
            dir_list = [[self.foldersSnap, 'Snapshots']]  # As a minimum we should have the snapshots directory
        self.image_options = [dir_info for dir_info in dir_list if os.path.isdir(dir_info[0])]
        self.my_logger.info(_('Completed Reading INI file settings'))  # Write to the log file

    # ----------------------------------------------------------------------------------------------------------------
    # Scroll the image option selection to the right
    def image_options_left(self, _event) -> None:
        if self.image_options_index > 0:
            self.image_options_index -= 1
            self.image_option.SetLabel(self.image_options[self.image_options_index][1])
            self.on_single_click(-1)

    # ----------------------------------------------------------------------------------------------------------------
    # Scroll the image option selection to the left
    def image_options_right(self, _event) -> None:
        if self.image_options_index < len(self.image_options) - 1:
            self.image_options_index += 1
            self.image_option.SetLabel(self.image_options[self.image_options_index][1])
            self.on_single_click(-1)

    # ------------------------------------------------------------------------------------------------------------------
    # This event captures a left mouse click on the URL at the end of the ROM info window.  It then launches
    # the default web browser pointing at that URL.
    def on_url_selected(self, event) -> None:
        self.SetCursor(wx.Cursor(wx.CURSOR_HAND))
        if event.MouseEvent.LeftUp():
            text = event.GetEventObject().GetValue()
            pos = text.find('http')
            url = text[pos:]
            webbrowser.open(url)
        event.Skip()

    # ------------------------------------------------------------------------------------------------------------------
    # Open the pfeMAME sourceforge web page in the default browser
    @staticmethod
    def open_web_page(_self) -> None:
        webbrowser.open('https://sourceforge.net/projects/pfemame/')

    # ------------------------------------------------------------------------------------------------------------------
    # Get the version information from sourceforge to see if there is a newer version of pfeMAME available.
    # This is done by looking for any files listed that match the operating system type in the following formats:
    # pfemame_windows_v
    # pfemame_linux_v
    # pfemame_mac_v
    # The text after the 'v' will be the version number.
    def check_for_updates(self, event) -> None:

        # This function is to be used with CallAfter to allow a message box to be displayed from a separate thread
        # in a thread safe manner.
        def display_threadsafe_message(message: tuple) -> None:
            if not message[3]:
                # Display a standard dialog message
                dlg = wx.MessageDialog(None,
                                       message[0],
                                       message[1],
                                       message[2])
                dlg.ShowModal()
                dlg.Destroy()
                self.Refresh()
            else:
                # Display a dialog message with the checkbox to skip it next time
                dlg = wx.RichMessageDialog(None,
                                           message[0],
                                           message[1],
                                           message[2])
                dlg.ShowCheckBox(_('Don\'t check software on startup'))
                dlg.ShowModal()
                if dlg.IsCheckBoxChecked():
                    # Update the database date in the INI file
                    iniFile.set_ini_file_variables([('preferences', 'checksoftwareversion', 'False')])
                    pub.sendMessage('disable_software_version_check')
                dlg.Destroy()
                self.Refresh()

        def run_update_thread(update_type: str) -> None:
            wx.CallAfter(self.statusbar_panel.set_status, _('Checking for Software Updates'))
            self.my_logger.info(_('Checking for software update - requesting data from sourceforge'))

            user_agent = 'Mozilla/5.0 (Windows NT 6.1; Win64; x64)'
            headers = {'User-Agent': user_agent}
            url = 'https://sourceforge.net/projects/pfemame/files'
            req = Request(url, data=None, headers=headers)
            try:
                context = ssl._create_unverified_context()  # Stop issues with SSL certificates
                content = str(urlopen(req, context=context).read())
            except urllib.error.URLError:
                wx.CallAfter(self.statusbar_panel.set_status, _('Ready'))
                msg = _('Checking for software update - unable to access network connection '
                      'or update version file not found')
                self.my_logger.error(msg)
                wx.CallAfter(self.messages_append, f'{msg}\n', True)
                wx.CallAfter(self.enable_alert_timer)
                return

            # Do a string match based on the OS type
            systems = {'windows': 17, 'linux': 15, 'mac': 13}
            length = systems[self.system]

            index = content.find(f'pfemame_{self.system}_v')
            online_version = content[index + length:index + content[index:].find('.7z')]  # Grab the version

            msg = _('Checking for software update - got data from sourceforge')
            self.my_logger.info(msg)

            try:
                self.my_logger.info(_('Checking for software update - trying to calculate version'))

                result = functions.compare_version(online_version, __version__)
                if result == 1:
                    self.my_logger.info(_('Checking for software update - new version exists'))
                    wx.CallAfter(display_threadsafe_message,
                                 (_('A newer version of pfeMAME is available at Sourceforge.  '
                                   'You have version {} while the latest version is {}').format(__version__, online_version),
                                  _('New Version Available'),
                                  wx.OK | wx.ICON_INFORMATION, True))
                else:
                    msg = _('Checking for software update - you already have the latest version')
                    self.my_logger.info(msg)
                    wx.CallAfter(self.messages_append, f'{msg}\n', True)
                    if update_type != 'auto':
                        wx.CallAfter(display_threadsafe_message,
                                     (_('You already have the latest version of pfeMAME installed.'),
                                      _('Check for software update'),
                                      wx.OK | wx.ICON_INFORMATION, False))
            except TypeError:
                self.my_logger.error(_('Unable to get version information from Sourceforge.  Request returned: {}. '
                                'It should have returned three numbers separated by a dot (e.g. 2.2.2)').format(online_version))
                if update_type != 'auto':
                    wx.CallAfter(display_threadsafe_message,
                                 (_('An error has occurred and I am unable to check the online version. '
                                  'Check the logfile for more information.'),
                                  _('Error retrieving version'),
                                  wx.OK | wx.ICON_INFORMATION, False))
            wx.CallAfter(self.statusbar_panel.set_status, _('Ready'))

        # Run the update checking function in a separate thread
        thread = Thread(target=run_update_thread, args=(event,))
        thread.daemon = True  # Force the thread to close with the main thread
        thread.start()

    # ------------------------------------------------------------------------------------------------------------------
    # Game categories double-clicked event
    # This will toggle all selected and none selected when you double-click in the ROM categories window
    def on_doubleclick_categories(self, _event) -> None:
        checkbox_controls = [widget for widget in self.panel1.GetChildren() if isinstance(widget, wx.CheckBox)]
        any_checked = next((x for x in checkbox_controls if x.GetValue() is True), False)
        for ctrl in checkbox_controls:
            ctrl.SetValue(not any_checked)
        self.on_update_categories(-1)

    # ------------------------------------------------------------------------------------------------------------------
    # ROM categories checkbox clicked event
    # This will update the categories.dat file with the current state of the ROM categories check boxes
    # and then refresh the ROMs list to show the new selection.
    def on_update_categories(self, _event) -> None:
        # Look at the state of each checkbox control, and get its name if it's checked.
        # The name was set to the unique categories so can be used to determine which
        # categories I want selected.
        # Write these out to the 'categories.dat' text file for later use.
        checkbox_controls = [widget for widget in self.panel1.GetChildren() if isinstance(widget, wx.CheckBox)]
        with open(os.path.join(path_to_me, 'categories.dat'), 'w') as fp:
            try:
                for ctrl in checkbox_controls:
                    if ctrl.GetValue() is True:
                        fp.write(ctrl.GetName() + '\n')
            except IOError:
                pass

        self.comboBoxFilter.SetValue(_('All'))
        self.controlSearch.SetValue('')  # Empty the search input box
        self.load_categories()  # Rebuild the categories list
        number_roms_in_filtered_list = self.populate_list(-1, '', False)  # Update the ROMs view
        self.update_rom_numbers(number_roms_in_filtered_list)

    # --------------------------------------------------------------------------------------------------------------
    # This is called by clicking on the 'Play Video' control and by the video timer event.  It makes sure that the
    # main ROMs list is given back focus.
    def on_play_clicked(self, _event) -> None:
        if self.use_vlc:
            self.vlc_playing = True
            # Play video from a thread
            thread = Thread(target=self.check_status, args=())
            # thread = Thread(target=self.on_play, args=())
            thread.daemon = True  # Force the thread to close with the main thread
            thread.start()
        else:
            # Load the video using MediaCtrl
            self.on_play()
        self.list.SetFocus()

    # Loop until the VLC video has ended, then force the media stop which will show the ROM image again
    def check_status(self) -> None:
        self.on_play()
        time.sleep(1)  # Just wait a bit for the video to be playing so that the following check works
        while self.vlc_playing and not self.close_selected:
            time.sleep(0.1)
        self.vlc_playing = False
        self.mc_forced_stop()

    # --------------------------------------------------------------------------------------------------------------
    # Load and play (if available) the ROM preview video.
    # This is in the same physical location as the ROM image.  One will be hidden to show the other.
    def on_play(self) -> None:
        if self.foldersVideo != '':
            if self.media_player_available:
                if self.use_vlc:
                    # This code is for playback using VLC
                    if self.player.is_playing() == 1:
                        self.mc_forced_stop()
                        return

                    # Ensure the correct position for the media control and show it
                    self.pnlVideo.SetPosition(self.image_panel.GetPosition())
                    # self.image_panel.Hide()
                    self.pnlVideo.Show()

                    # Play the video selected during the single click event
                    self.videoTimer.Stop()
                    try:
                        self.media = self.instance.media_new(self.video_to_play)
                        self.player.set_media(self.media)
                        self.player.play()
                        self.videoLoadCaption.SetLabel(_('Playing Video - Click To Stop'))
                        # The following code is a fudge fix for macOS.  For some reason on macOS the VLC video
                        # doesn't automatically resize to fit the window.  Some reports have indicated that this
                        # is a version issue of either macOS or VLC, but I have not found anything concrete on this.
                        # To work around the issue I pause a bit so that the video is running, then get the
                        # video vertical size and scale the video to the maximum that can be fitted onto the screen.

                        if self.system == 'mac':
                            # Wait until the video is playing so we can get the size
                            while self.player.is_playing() == 0:
                                time.sleep(0.1)

                            video_height = self.player.video_get_height()
                            screen_height = self.image_size_y
                            if video_height != 0:  # Just in case, don't try to divide by zero
                                screen_scale = screen_height / video_height
                                self.player.video_set_scale(screen_scale)

                                # self.player.video_set_scale(2.0)
                            self.player.stop()
                            self.player.play()

                    except wx._core.wxAssertionError:
                        msg = _('Unable to play video on this platform.')
                        self.my_logger.error(msg)
                        self.media_player_available = False
                else:
                    # This code is for playback using wx MediaCtrl
                    if self.mc.GetState() == 2:
                        # Media player is currently playing, so stop it
                        self.mc_forced_stop()
                        return

                    # Ensure the correct position for the media control and show it
                    self.mc.SetPosition(self.image_panel.GetPosition())
                    self.image_panel.Hide()
                    self.mc.Show()

                    # Play the video selected during the single click event
                    self.videoTimer.Stop()
                    try:
                        if not self.mc.Load(self.video_to_play):  # Load the video
                            self.messages_append(_('Unable to load {}: Unsupported format?\n').format(self.video_to_play))
                    except wx._core.wxAssertionError:
                        msg = _('Unable to play video on this platform.')
                        self.my_logger.error(msg)
                        self.media_player_available = False
        else:
            msg = _('Tried playing ROM video but the path to the videos folder is not set.')
            self.my_logger.error(msg)
            self.messages_append(f'ERROR: {msg}\n')

    def inactivity_timer_start(self) -> None:
        if not self.inactivity_timer_block and not self.video_disable and self.prefScreenSaverTime != '0':
            self.inactivity_timer.Start(int(self.prefScreenSaverTime) * 1000 * 60)

    # A timer event has been triggered
    def on_timer_event(self, event) -> None:
        event_id = event.GetId()
        if event_id == 9991:
            # It's the alert timer that's triggered this event
            # Starting at '#FFF0F0'
            # Decrement the value of the 2nd & 3rd pair by 10h (16D) at a time until we reach '#FF0000'
            # Then increment in the same steps back to the start
            alert_color = ''.join(['#ff', '{:02x}'.format(self.alert_colors_value),
                                   '{:02x}'.format(self.alert_colors_value)])
            self.controlMessages.SetBackgroundColour(alert_color)
            self.controlMessages.Refresh()
            self.alert_colors_value -= self.alert_colors_decrement_value  # Decrement / increment to next value
            # If we have hit the bottom or top of the range then invert the incrementing direction
            if self.alert_colors_value == -16 or self.alert_colors_value == 256:
                self.alert_colors_decrement_value = -self.alert_colors_decrement_value
                self.alert_colors_value = 0 if self.alert_colors_value == -16 else 240
        elif event_id == 9992:
            # It's the inactivity timer, so we need to stop the timer and open the screensaver
            self.on_fullscreen_video_player(-2)
        else:
            # It must be a video timer event
            self.timerSeconds -= 1
            self.videoLoadCaption.SetLabel(_('Auto loading video in {} seconds').format(str(self.timerSeconds)))
            if self.timerSeconds == 0:
                self.timerSeconds = 5
                self.on_play_clicked(-1)

    def on_media_loaded(self, _event) -> None:
        self.mc.Play()  # Play the video
        self.videoLoadCaption.SetLabel(_('Playing Video - Click To Stop'))

    def on_media_stopped(self, _event) -> None:
        self.vlc_playing = False
        self.videoLoadCaption.SetLabel(_('Video Stopped'))
        self.on_single_click(-1)  # Show the ROM image as the video has now stopped

    def on_media_stopped_vlc(self, _event) -> None:
        self.vlc_playing = False
        self.videoLoadCaption.SetLabel(_('Video Stopped'))

    # Force the ROM video player to stop
    def mc_forced_stop(self):  # Stop the ROM video player
        try:
            # Stop any playing video and hide the control
            if self.media_player_available:
                self.videoLoadCaption.SetLabel(_('Video Stopped'))
                if self.use_vlc:
                    self.player.stop()
                    self.pnlVideo.Hide()
                else:
                    self.mc.Stop()
                    self.mc.Hide()
                self.image_panel.Show()
                self.videoTimer.Stop()  # Stop the video timer
                self.timerSeconds = 5  # Set the timer count back to 5
        except AttributeError:
            pass

    # ------------------------------------------------------------------------------------------------------------------
    # Show the full screen video player
    def on_fullscreen_video_player(self, _event) -> None:
        self.inactivity_timer.Stop()  # Stop the inactivity timer
        self.video_stop()  # Stop the video

        video_file_types = ['.mpeg', '.avi', '.asf', '.wmv', '.mp4', '.mov', '.3gp', '.ogg', '.ogm',
                                 '.mk4', '.m4v']
        video_list = [f for f in os.listdir(self.foldersVideo) if (os.path.isfile(os.path.join(self.foldersVideo, f))
                                                                      and (os.path.splitext(f)[1] in
                                                                           video_file_types))]

        # Only do these checks if we have manually requested to show the player
        if _event != -2:
            # Check if we have a valid Videos folder set
            if not os.path.exists(self.foldersVideo) or self.foldersVideo == '.':
                msg = _('Please set up a valid Videos folder in Filepaths')
                dlg = wx.MessageDialog(self,
                                       msg,
                                       _('ERROR'),
                                       style=wx.OK)
                dlg.ShowModal()
                return

            # If no videos are found then exit
            if len(video_list) == 0:
                msg = _("It doesn't look like there are any valid video files in the Video folder")
                dlg = wx.MessageDialog(self,
                                       msg,
                                       _('ERROR'),
                                       style=wx.OK)
                dlg.ShowModal()
                return

        if os.path.exists(self.foldersVideo) and self.foldersVideo != '.' and not self.video_disable:
            fullscreenplay.FullScreenVideoPlayer(self.foldersVideo, 'mainview', self, self.force_disable_vlc,
                                                 self.force_video)

    # ------------------------------------------------------------------------------------------------------------------
    # Clear the messages text box
    def on_clear_messages(self, _event) -> None:
        self.controlMessages.Clear()

    # ------------------------------------------------------------------------------------------------------------------
    # Copy the messages window content to the clipboard
    def on_copy_messages(self, _event) -> None:
        data_object = wx.TextDataObject()
        data_object.SetText(self.controlMessages.GetValue())
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(data_object)
            wx.TheClipboard.Close()
        else:
            wx.MessageBox(_('Unable to open the clipboard'), _('Error'))

    def on_enable_alert_timer(self, _event):
        self.enable_alert_timer()

    # ------------------------------------------------------------------------------------------------------------------
    # Load the history file into memory
    def load_history_dat(self) -> None:
        self.history_data = functions.load_history_dat(self, self.my_logger, self.pathHistory, 'history')

    # ------------------------------------------------------------------------------------------------------------------
    # Load the mameinfo.dat file into memory
    def load_mameinfo_dat(self) -> None:
        self.mameinfo_data = functions.load_history_dat(self, self.my_logger, self.pathMAMEInfoDat, 'mameinfo')

    # ------------------------------------------------------------------------------------------------------------------
    # Add the ROM specific preferences tab to the main list view.
    # If it already exists then delete it and recreate.  This is done so that the tab changes with changed
    # roms.
    def preferences_rom_specific(self, _event) -> None:

        if os.path.exists(self.folder_ini):
            # self.Freeze()
            results = self.get_selected_row_rom_file_name()
            rom_file_name = results['rom_file_name']
            rom_name = results['rom_name']
            system_file_name = results['system_file_name']
            category = results['rom_type']

            is_a_system = True if category in systems_list else False

            # If it's a system program then add that to the title
            if system_file_name != '':
                rom_name = f'{rom_name}: {system_file_name}'

            # If we have more than two pages then delete the third page as it's the old ROM specific preferences
            # and we need to remove that before we add the new one.
            count = self.list_notebook.GetPageCount()
            if count > 2:
                self.list_notebook.DeletePage(2)

            if rom_name != '':  # Don't show if we don't have a valid ROM selected
                self.list_notebook.AddPage(preferences.Preferences(parent=self.list_notebook,
                                                                   title=_('ROM Preferences - {}').format(rom_name),
                                                                   mode='game',
                                                                   rom_file_name=rom_file_name,
                                                                   folder_ini=self.folder_ini,
                                                                   path_executable=self.path_mame_exe,
                                                                   system_file_name=system_file_name,
                                                                   is_a_system=is_a_system),
                                           _('{} Preferences').format(rom_name))

            # Show the ROM specific preferences tab
            self.list_notebook.SetSelection(2)
            # self.Thaw()
        else:
            dlg = wx.MessageDialog(self, 
                                   _('The INI Folder path is not valid.  Please select EDIT > File Paths and set a '
                                   'valid path for this folder.'),
                                   _('No INI folder path set'),
                                   wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
            dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # Process the tile view menu option event
    def on_tile_view(self, _event) -> None:
        self.video_stop()  # Stop the video
        self.inactivity_timer.Stop()  # Stop the inactivity timer from this view
        self.tileview_running = True  # Flag to prevent joystick events from doing anything when tileview is running
        self.my_logger.info(_('Switching to tile view'))  # Write to the log file
        tileview.TileView(mylist_local=self.mylist, 
                          my_logger=self.my_logger,
                          version=__version__, 
                          mame_version=self.mame_version,
                          copy_right=__copyright__,
                          video_disable=self.video_disable,
                          disable_vlc=self.force_disable_vlc,
                          force_video=self.force_video)

    # ------------------------------------------------------------------------------------------------------------------
    # This function just allows a simple way for child windows to call the
    # log file writing custom event remotely.
    # def write_log_remote(self, message) -> None:
    #     my_logger.info(message)  # Write to the log file

    # ------------------------------------------------------------------------------------------------------------------
    # When requested, rebuild the full ROMs database (refresh and reload)
    def on_database_refresh(self, event) -> None:
        # Load the categories file just to make sure it's correct.  This has to be done as during first time run
        # the categories file can't be found.  Once you have set up the paths you just need to refresh the database
        # and it will be loaded here (Rather than having to restart the application).
        if not self.categories:
            self.load_categories()  # Rebuild the categories list

        self.list.Disable()  # Disable the ROMs list to prevent errors with clicking while it is being refreshed
        self.video_stop()  # Stop the video

        # This checks to see if the MAME executable file is where it should be.  It's a basic check just to try and
        # prevent odd things from occurring if the file is missing - maybe an upgrade was done.
        if not os.path.exists(self.path_mame_exe) and self.path_mame_exe != '' and self.path_mame_exe != 'mame':
            message = _('Unable to find the MAME executable.  If this is the first time run then ignore this warning.  '
                      'If not, please check the File Paths window for the correct path to the MAME executable file.  '
                      'Recent updates to MAME 64 bit have removed the old 64 extension to the file name.')
            dlg = wx.MessageDialog(self, message, 'ERROR', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
            dlg.Destroy()
            return

        self.availableRoms = 0  # Just a fudge for now
        self.mylist = []

        def do_full_refresh() -> int:
            # Perform a full rebuild of the ROMs database
            self.my_logger.info(_('Started building full ROMs database'))  # Write to the log file
            start = datetime.datetime.now()
            # Build up the full ROMs list database into 'mylist' along with default values for flags
            self.build_rom_list(self.category_dict)
            if len(self.mylist) == 0:
                return 0

            # Check if each ROM in the full MAME list actually exists in the rom path
            self.availableRoms = self.get_exists()
            # Check each ROM in the favorites file and set the favorites flag in mylist
            self.get_dat_file_data('favorites.dat')
            # Check each ROM in the ratings file and set its star rating
            self.get_dat_file_data('ratings.dat')
            # Check each ROM in the runs file and update its number of runs indicator
            self.get_dat_file_data('runs.dat')
            iniFile.set_ini_file_variables(
                [('system', 'cachedate', self.today)])  # Update the database date in the INI file
            msg = _('ROMs database refresh completed in '
                  '{} seconds.').format(str((datetime.datetime.now() - start).seconds))  # Calculate the execution time

            self.my_logger.info(msg)
            self.messages_append(f'{msg}\n', True)
            self.write_games_database()  # Write out the ROMs database (mylist) to a cache file

            return self.availableRoms

        if self.path_mame_exe != '' and self.folders_roms != '':
            if event == 'auto' and self.cacheDate != '':
                # Read the temporary database file in
                self.my_logger.info(_('Started reading the temporary ROMs database file'))  # Write to the log file

                try:
                    # Try to load in the database cached version of mylist
                    with shelve.open(os.path.join(path_to_me, 'mylist.db')) as mylist_db:
                        self.mylist = list(mylist_db.values())[0]

                    # Check the number of columns of data in the loaded database.  If it doesn't match the current
                    # design then force an exception to rebuild from scratch.  This will manage updates that I 
                    # make to the database design when adding more data to the list view.
                    if len(self.mylist[0]) != display_columns:
                        msg = _('ROMs database columns count mismatch.  Forcing full rebuild.')
                        self.my_logger.error(msg)
                        self.messages_append(f'{msg}\n', True)
                        raise Exception

                    self.availableRoms = sum(int(row[index_exists]) for row in self.mylist)  # Update the available roms

                    self.my_logger.info(_('Completed reading the temporary ROMs database file'))  # Write to the log file
                except (IndexError, FileNotFoundError, KeyError, Exception):
                    # Can't load the database file, so we need to force a full database refresh
                    msg = _('Unable to read the temporary ROMs database file - recreating from scratch')
                    self.my_logger.error(msg)
                    self.messages_append(f'{msg}\n', True)
                    self.availableRoms = do_full_refresh()
                    self.enable_alert_timer()
            else:
                # We are rebuilding the full ROMs database from scratch
                self.availableRoms = do_full_refresh()

            if self.start_up_complete is True:
                # Populate the ROMs list showing all ROMs as default
                number_roms_in_filtered_list = self.populate_list(filtercolumn=-1,
                                                                  filtervalue='',
                                                                  read_from_cache=False)

                self.update_rom_numbers(number_roms_in_filtered_list)
                if number_roms_in_filtered_list == 0:
                    dlg = wx.MessageDialog(self,
                                           _('Looks like your ROMs list is empty.  '
                                           'Please check that you have at least one ROM category selected.'),
                                           _('Empty ROMs list'),
                                           wx.OK | wx.ICON_INFORMATION)
                    dlg.ShowModal()
                    dlg.Destroy()

            self.comboBoxFilter.SetValue('All')

            # Empty the search input box - we have to prevent the search input event from triggering which happens
            # when you write anything to the control.  If we don't prevent this then the games list gets emptied.
            self.controlSearch.Enable(False)
            self.clear_event_override = True  # Prevent the search input event from triggering
            self.controlSearch.SetValue('')  # Empty the search input box
            self.controlSearch.Enable()
            self.clear_event_override = False

            self.text_database_ageValue.SetLabel(_('0 days'))  # Update the games database cache age display
            self.statusbar_panel.set_status(_('Ready'))

            # Load the history file just to make sure it's correct.  This has to be done as during first time run
            # the history file can't be found.  Once you have set up the paths you just need to refresh the database
            # and it will be loaded here (Rather than having to restart the application).
            try:
                if not self.history_data:
                    self.load_history_dat()
            except NameError:
                self.load_history_dat()
        else:
            # This is most likely the first time pfeMAME has been run so provide the user some prompts
            self.statusbar_panel.set_status(_('File Paths need to be set up'))
            self.Refresh()
            # Highlight the background color of the File Paths toolbar icon to make it obvious this is where to go
            for widget in self.panel.GetChildren():
                if widget.GetName() == 'File Paths':
                    widget.SetBackgroundColour(wx.RED)
                    widget.Refresh()
                    break

            # Show the language picker and then reload the language if it has not yet been set
            if iniFile.read_ini_file('preferences', 'language') == '':
                dlg = languagepicker.LanguagePicker()
                dlg.ShowModal()
                dlg.Destroy()

            dlg = wx.MessageDialog(self,
                                   _('It Looks like this is the first time you have run pfeMAME'
                                   ' as the INI file data is missing.  Your first step is to '
                                   'set up the necessary file paths to MAME files and supporting'
                                   ' folders. \n \nPlease open EDIT > File Paths and set the MAME '
                                   'executable file path and ROMS folder as a minimum.  '
                                   'NOTE:  You can use the Auto Find option in the file paths '
                                   'window to make this a lot easier in most cases.  \n \nAfter this '
                                   'you need to select some ROM categories (the right of the screen).'
                                   '\n \nThen restart the application and refresh the ROMs database '
                                   '\nVIEW > Refresh ROMs Database.'),
                                   _('MAME file path or ROMS folder not found'),
                                   wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
 
            # Force all ROM categories to be selected for first-time-run
            checkbox_controls = [widget for widget in self.panel1.GetChildren() if isinstance(widget, wx.CheckBox)]
            for ctrl in checkbox_controls:
                ctrl.SetValue(True)
            self.on_update_categories(-1)

        self.list.Enable()
        # Give the ROMs list the focus
        self.list.SetFocus()

    # ------------------------------------------------------------------------------------------------------------------
    # Write out the games database (mylist) to a cache file.  This is then reloaded on
    # startup to speed the startup process.
    def write_games_database(self) -> None:
        # Write the temporary mylist file out - this can be read in later to speed things up.
        self.my_logger.info(_('Writing the temporary ROMs database file'))  # Write to the log file
        try:
            with shelve.open(os.path.join(path_to_me, 'mylist.db'), flag='n') as mylist_db:
                mylist_db['mylist'] = self.mylist
        except IOError:
            self.my_logger.error(_('Unable to write the temporary ROMs database file'))  # Write to the log file

    # ------------------------------------------------------------------------------------------------------------------
    # Build a list with game name and category from the catver.ini file
    # This is later used when building the game list to find out what the games category is
    # category_dict is the dictionary created
    def build_categories(self) -> (dict, str):
        self.statusbar_panel.set_status(_('Building ROM & category list.  Please Wait.........'))
        self.my_logger.info(_('Reading categories from catver.ini'))  # Write to the log file

        category_dict = {}
        catver_version = ''
        try:
            with open(self.pathCatver, 'r') as fp:
                lines = fp.readlines()

            # Get the catver.ini file version
            for line in lines:
                if ';; catver.ini' in line.lower():
                    start = line.find('MAME')
                    catver_version = line[start + 5: -4]
                    self.my_logger.info(_('catver.ini file version: {}').format(catver_version))  # Write to the log file
                    break

            category_start = lines.index('[Category]\n')
            category_end = lines.index('[VerAdded]\n')
            lines = lines[category_start + 1:category_end - 1]

            category_dict = {i.partition('=')[0].rstrip(): i.partition('=')[2].partition('/')[0].rstrip()
                             for i in lines}

            # The following are added to allow games that have no classification as well as pinball
            # games (VisualPinball) to be included.
            category_dict['.1'] = 'Not Classified'
            category_dict['.2'] = 'Pinball'

        except ValueError:
            dlg = wx.MessageDialog(self,
                                   _('Parsing error in the catver.ini file.  Please revert to a previous version.'),
                                   _('ERROR'),
                                   wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
        except PermissionError:
            dlg = wx.MessageDialog(self,
                                   _('Unable to read the catver.ini file.  Please check file and folder permissions.'),
                                   _('ERROR'),
                                   wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
        except FileNotFoundError:
            dlg = wx.MessageDialog(self,
                                   _('Cannot find the catver.ini file.  Please check file path settings.'),
                                   _('ERROR'),
                                   wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()

        self.my_logger.info(_('Completed reading categories from catver.ini'))  # Write to the log file
        return category_dict, catver_version

    # ------------------------------------------------------------------------------------------------------------------
    # Build up the ROMs list database
    # This section calls the MAME command line application with the '-listxml' option
    # The '-listxml' option generates an output of each rom name and corresponding game name
    # A list is then built (self.mylist) based on the rom description and file name.
    # The list is structured as follows;
    # [0] - rom file name
    # [1] - rom name (description)
    # [2] - system file name if applicable (e.g. spectrum)
    # [3] - rom exists flag
    # [4] - rom rating
    # [5] - number of runs
    # [6] - favorite flag
    # [7] - rom category (shooter etc)
    # [8] - system short name taken from the softlist (-listsoftware)
    # [9] - system programs folder name (e.g. spectrum_cass)

    def build_rom_list(self, category_dict: dict) -> None:
        self.mylist = []

        # Only do this if there is a valid path to MAME, otherwise we will crash
        if os.path.isfile(self.path_mame_exe) or self.path_mame_exe == 'mame':
            self.statusbar_panel.set_status(_('Building ROM list.  Please Wait.........'))
            self.my_logger.info(
                _('Building full ROM list using MAME -listxml command and assigning category into mylist'))
            if self.pref_full_systems_method is True:
                self.my_logger.info(_('> Using the full systems search method'))
            else:
                self.my_logger.info(_('> Using the basic systems search method'))

            wx.Yield()  # Make sure the status bar message is displayed before the long-running task below starts

            # Check to see if the XML Cache file exists.  If it does, read it in as that's a lot faster than
            # running the mame.exe -listxml command.
            data = []
            cache_read_ok = False
            self.my_logger.info(_('Trying to read the XML cache file'))
            if os.path.exists(os.path.join(self.path_to_me, 'xmlcache.dat')):
                with open(os.path.join(self.path_to_me, 'xmlcache.dat'), 'r', encoding='utf-8') as fp:
                    data = fp.readlines()

                # Check the version info in the cache file
                if len(data) > 0:
                    for index, line in enumerate(data):
                        if '<mame build="' in line:
                            cache_version = line.split('<mame build="')[1].split(' ')[0]
                            if functions.compare_version(cache_version, self.mame_version) == 0:
                                cache_read_ok = True
                            break
            else:
                self.my_logger.info(_('XML Cache file does not exist'))

            # If we haven't read out the XML cache file, or if the MAME version has changed since it was last written,
            # then we will need to perform the mame.exe -listxml command which can take a while
            if not cache_read_ok:
                # Execute the MAME - listxml command using the full file path to the MAME executable
                self.my_logger.info(_('Calling mame.exe -listxml to get mame program data'))
                t1 = busywindow.PleaseWait()  # Show the busy window
                wx.Yield()
                cmd = [self.path_mame_exe, '-listxml', '-nodtd']
                output = subprocess.Popen(cmd,
                                          stdout=subprocess.PIPE,
                                          shell=self.shell_to_use,
                                          universal_newlines=True,
                                          encoding='utf-8')
                data = output.stdout.readlines()
                if len(data) == 0:
                    self.my_logger.error(_('ERROR: I could not get a result from the mame.exe -listxml command.  '
                                    'Check that file paths are correct.'))
                    t1.on_close(-1)  # Close the busy window
                    wx.Yield()
                    return

                # Write out the xml data to a cache file.  We can reload that again to speed up the process assuming the
                # MAME version has not changed.
                try:
                    with open(os.path.join(self.path_to_me, 'xmlcache.dat'), 'w', encoding='utf-8') as fp:
                        fp.writelines(data)
                except IOError:
                    self.my_logger.error(_('ERROR: Unable to write the XML Cache file after listxml.  '
                                    'Check folder permissions for the pfeMAME folder.'))

                self.my_logger.info(_('XML Cache File written'))

                t1.on_close(-1)  # Close the busy window
                wx.Yield()

            # Replace '&' escaped characters in the output data
            data = functions.replace_escaped_characters(data)

            # Parse the info retrieved from MAME
            lines = []
            for index, line in enumerate(data):
                if '<machine name=' in line and 'isdevice="yes"' not in line:
                    machine_name = line.split('machine name="')[1].split('"')[0]
                    description = data[index + 1].split('<description>')[1].split('</description>')[0]
                    year = data[index + 2].split('<year>')[1].split('</year>')[0]
                    manufacturer = data[index + 3].split('<manufacturer>')[1].split('</manufacturer>')[0]
                    lines.append([machine_name, description, year, manufacturer])

            # Set the progress bar at its starting position, show it, and set its range to be the number of lines
            # in the ROMs list.
            self.progress_bar.SetValue(0)
            self.progress_bar.Show()
            self.progress_bar.SetRange(len(lines))

            # Need to save the current working directory, so it can be restored later, and change the current
            # directory to the location of the MAME executable
            old_path = path_to_me
            if self.system != 'linux':
                os.chdir(os.path.dirname(self.path_mame_exe))

            # Get a list of all files and folders in the ROMs (systems folders) path.  This will later be used to
            # locate systems programs.
            file_dictionary = {}
            try:
                for directory in set([x[0] for x in os.walk(self.folders_systems)]):
                    files = os.listdir(directory)
                    file_dictionary[directory] = files
                if self.pref_use_multiprocessor == 'True':
                    manager = multiprocessing.Manager()
                    file_dictionary = manager.dict(file_dictionary)
            except StopIteration:
                pass

            def get_category(__) -> (str, str, str, str, str):
                # Try to get the ROM category for this rom name.  If not available then make it 'Not Classified'
                # so that at least it can be displayed in the ROMs list.
                # Get the separate rom name and full ROM name from the list
                rom_file_name = __[0]
                game_name = __[1]
                year = __[2]
                manufacturer = __[3]
                category = category_dict.get(rom_file_name)
                # If the rom name doesn't exist in the category dictionary, then the lookup above will return a
                # NoneType.  This would be bad for my code, so lets catch it and substitute that situation with a
                # 'Not Classified' in the category dictionary.
                if category is None:
                    category = 'Not Classified'
                return category, rom_file_name, game_name, year, manufacturer

            # Create some local variables
            self.mylist = []
            extend = self.mylist.extend
            pref_full_systems_method = self.pref_full_systems_method
            path_mame_exe = self.path_mame_exe
            shell_to_use = self.shell_to_use
            folders_systems = self.folders_systems
            errors = []

            # This function is just the callback helper for the multiprocessor call below.
            # It just builds up the result which will be processed later.
            def callback(result: list[str]) -> None:
                self.mylist.extend(result[0])
                if result[1] is not None:
                    errors.append(result[1])

            # This is the multiprocessor version of the ROMs database build code
            if self.pref_use_multiprocessor == 'True':
                pool = multiprocessing.Pool(processes=multiprocessing.cpu_count() - 1)
                processes = []
                for line in lines:
                    category, rom_name, game_name, year, manufacturer = get_category(line)
                    processes.append(pool.apply_async(database_build.get_info, (self.prefSystemsSupport,
                                                                                pref_full_systems_method,
                                                                                category,
                                                                                game_name,
                                                                                rom_name,
                                                                                path_mame_exe,
                                                                                shell_to_use,
                                                                                folders_systems,
                                                                                systems_list,
                                                                                file_dictionary,
                                                                                year,
                                                                                manufacturer,), callback=callback))
                # Wait for each process to complete - as each one does, update the progress bar and refresh the
                # window using Yield so that we don't lock up the main GUI.
                for index_count, process in enumerate(processes):
                    self.progress_bar.SetValue(self.progress_bar.GetValue() + 1)  # Update the progress bar
                    if not (index_count % 100):
                        self.progress_bar.Refresh()
                        wx.GetApp().Yield()
                    process.wait()

                # Make sure the pool is closed and all child processes are stopped
                pool.close()
                pool.join()
            else:
                # This is the Non multiprocessor version of the ROMs database build code
                for index_count, line in enumerate(lines):
                    category, rom_name, game_name, year, manufacturer = get_category(line)
                    result = database_build.get_info(self.prefSystemsSupport,
                                                     pref_full_systems_method,
                                                     category,
                                                     game_name,
                                                     rom_name,
                                                     path_mame_exe,
                                                     shell_to_use,
                                                     folders_systems,
                                                     systems_list,
                                                     file_dictionary,
                                                     year,
                                                     manufacturer,)
                    extend(result[0])
                    if result[1] is not None:
                        errors.append(result[1])
                    self.progress_bar.SetValue(self.progress_bar.GetValue() + 1)  # Update the progress bar
                    if not (index_count % 100):
                        self.progress_bar.Refresh()
                        wx.GetApp().Yield()

            os.chdir(old_path)  # Set the current working directory back

            # Check for additional non-MAME emulators that have a specified ROMs folder.  These are added to the
            # ROMs list the same way that system programs are.  The emulator is listed first (e.g. Visual Pinball),
            # and then each ROM is listed after it.
            if len(self.alternate_emulation_engines) > 0 and not self.alternate_emulation_engines_error:
                try:
                    engine_old = ''
                    for line in self.alternate_emulation_engines:
                        engine = line[0]
                        # path_alternate_emulator_engine = line[1]
                        alternate_emulator_rom_folder = line[2]
                        # emulator_name = line[3]
                        emulator_description = line[3]
                        # use_this_engine = line[4]
                        # include_extension = line[5]
                        # include_folder_path = line[6]
                        # extra_commands = line[7]
                        extensions = line[8].split()
                        if alternate_emulator_rom_folder != 'Default' and engine_old != engine:
                            engine_old = engine  # Make sure we don't add duplicate rows
                            count = 0
                            if os.path.exists(alternate_emulator_rom_folder):
                                for count, file_name in enumerate(os.listdir(alternate_emulator_rom_folder)):
                                    # Only add the files that have a valid extension from the XMl file
                                    if os.path.splitext(file_name)[1].lower()[1:] in extensions:
                                        # Create the ROM data line for this emulator engine
                                        self.mylist.append([engine, engine,
                                                            os.path.splitext(file_name)[0], '1', '0', '0', '0',
                                                            emulator_description, '', '', '', ''])
                                self.my_logger.info(_('Added {} {} files to ROMs list').format(count, engine))

                except IndexError:
                    msg = _('ERROR:  There was a problem processing the non_mame_emulators.xml file.  There is likely an'
                          ' error in this file.  Please check it out.')
                    wx.CallAfter(pub.sendMessage, 'update', msg=f'{msg}\r\n')
                    msg = _('XML File Error : non_mame_emulators.xml file may be corrupted')
                    wx.CallAfter(pub.sendMessage, 'log', msg=msg)
                    wx.CallAfter(pub.sendMessage, 'enable_alert_timer')

            # If any errors were detected during the database build then manage them here
            if errors:
                errors = [str(x) for x in errors]  # Ensure they are all STR rather than Unicode etc
                self.my_logger.error(_('Errors detected when building ROMs database.  Check the build_errors.txt file.'))
                try:
                    with open(os.path.join(path_to_me, 'build_errors.txt'), 'w') as fp:
                        fp.writelines(errors)
                except IOError:
                    self.my_logger.error(_('Unable to write to the build_errors.txt file'))  # Write to the log file

                _dlg = wx.MessageDialog(self,
                                        _('Errors were detected when building the ROMs database.  Please check the '
                                        'build_errors.txt file for information.'),
                                        _('Errors during Database Build'),
                                        wx.OK | wx.ICON_INFORMATION)
                _dlg.ShowModal()
                _dlg.Destroy()

            # Sort the ROM list by the rom name (in lower case) so things display in the right order
            self.mylist.sort(key=lambda x: (x[index_rom_name].lower(), x[index_rom_name]))
            self.my_logger.info(_('Completed building full ROMs list'))  # Write to the log file
            self.progress_bar.SetValue(0)  # Reset the progress bar back to zero
            self.progress_bar.Hide()  # Hide the progress bar

    # ------------------------------------------------------------------------------------------------------------------
    # This function checks to see if each ROM file actually exists in the
    # rom file path.  If it does, we then set the 'exists' flag to 1.
    def get_exists(self) -> int:
        self.statusbar_panel.set_status(_('Setting ROM exists flags.  Please Wait.........'))
        self.Refresh()
        self.my_logger.info(_('Setting ROM exists flags'))  # Write to the log file
        for folderRoms in self.folders_roms.split(';'):
            if os.path.exists(folderRoms):
                # Get the list of files within the folder
                file_exists_list = set([os.path.splitext(s.lower())[0] for s in os.listdir(folderRoms)])

                # Now check each ROM in mylist and see if it exists
                self.progress_bar.SetValue(0)  # Set the progress bar at its starting position
                self.progress_bar.Show()  # Show the progress bar
                # Set the range of the progress bar (number of lines in the game list)
                self.progress_bar.SetRange(len(self.mylist))

                ml = self.mylist
                pb = self.progress_bar.SetValue

                for index, line in enumerate(ml):
                    if line[index_rom_file_name] in file_exists_list:
                        # We have a match, so we need to change the 'exists' flag in the nested list
                        ml[index][index_exists] = '1'
                    pb(index)  # Update the progress bar
                    # Refresh the progress bar every 100th increment
                    if not (index % 100):
                        self.progress_bar.Refresh()
                        wx.GetApp().Yield()
        self.my_logger.info(_('Completed setting ROM exists flags'))  # Write to the log file
        self.progress_bar.SetValue(0)  # Reset the progressbar back to zero
        self.progress_bar.Hide()  # Hide the progress bar
        return sum(int(row[index_exists]) for row in self.mylist)  # Number of available roms

    # ------------------------------------------------------------------------------------------------------------------
    # Select the preferences tab (used when pressing CTRL-P)
    def on_preferences_default(self) -> None:
        self.list_notebook.SetSelection(1)

    # ------------------------------------------------------------------------------------------------------------------
    # This function is called on a key event from the list control.
    # We have to handle the scroll up and down ourselves to get around a problem trying to use EVT_KEY_UP
    # If we are in the 'favorites' view, then we will delete the selected game from the favorites list (if its delete)
    # We are also checking for ENTER key on selected rom to run that rom
    def on_key_press(self, event) -> None:
        key_code = event.GetKeyCode()  # Get the key code of key pressed

        self.inactivity_timer_start()  # Reset the inactivity timer

        # Page down the list
        if key_code == wx.WXK_PAGEDOWN:
            self.scroll_list_down_page()

        # Page up the list
        if key_code == wx.WXK_PAGEUP:
            self.scroll_list_up_page()

        # Scroll the list down one line
        if key_code == wx.WXK_DOWN:
            self.scroll_list_down()

        # Scroll the list up one line
        if key_code == wx.WXK_UP:
            self.scroll_list_up()

        # If we have pressed the ENTER key, then run the selected rom
        if key_code == wx.WXK_RETURN or key_code == wx.WXK_NUMPAD_ENTER:
            self.on_double_click(ignore_columns=True)

        # If we have pressed delete (or backspace) then remove the selected game
        # from the favorites list
        if key_code == wx.WXK_DELETE or key_code == wx.WXK_BACK:
            self.delete_favorite()

        control_down = event.ControlDown()
        shift_down = event.ShiftDown()

        # Check for CTRL-F
        if key_code == 70 and control_down:
            self.on_find(-1)

        # Check for CTRL-P
        if key_code == 80 and control_down and not shift_down:
            self.on_preferences_default()

        # Check for SHIFT-CTRL-P
        if key_code == 80 and control_down and shift_down:
            self.preferences_rom_specific(-1)

        # Check for SHIFT-V
        if key_code == 86 and not control_down and shift_down:
            self.on_fullscreen_video_player(-1)

        # Check for F5
        if key_code == 344:
            self.on_database_refresh(-1)

        # Check for CTRL-Q
        if key_code == 81 and control_down:
            self.on_close(-1)

    # ------------------------------------------------------------------------------------------------------------------
    # Scroll the list down one page
    def scroll_list_down_page(self) -> None:
        self.mc_forced_stop()  # Stop the video player and reset the timer
        count_per_page = self.list.GetCountPerPage()
        current_row = self.list.GetFirstSelected()
        try:
            self.list.Select(current_row + count_per_page)
            self.list.EnsureVisible(current_row + count_per_page)
        except Exception:
            # We have likely hit the bottom so don't go any further
            pass

    # ------------------------------------------------------------------------------------------------------------------
    # Scroll the list up one page
    def scroll_list_up_page(self) -> None:
        self.mc_forced_stop()  # Stop the video player and reset the timer
        count_per_page = self.list.GetCountPerPage()
        current_row = self.list.GetFirstSelected()
        try:
            self.list.Select(current_row - count_per_page)
            self.list.EnsureVisible(current_row - count_per_page)
        except Exception:
            # We have likely hit the top so don't go any further
            pass

    # ------------------------------------------------------------------------------------------------------------------
    # Scroll the list down one line
    def scroll_list_down(self) -> None:
        self.mc_forced_stop()  # Stop the video player and reset the timer
        current_row = self.list.GetFirstSelected()
        if current_row < self.list.GetItemCount() - 1:
            self.list.Select(current_row + 1)
            self.list.EnsureVisible(current_row + 1)

    # ------------------------------------------------------------------------------------------------------------------
    # Scroll the list up one line
    def scroll_list_up(self) -> None:
        self.mc_forced_stop()  # Stop the video player and reset the timer
        current_row = self.list.GetFirstSelected()
        if current_row > 0:
            self.list.Select(current_row - 1)
            self.list.EnsureVisible(current_row - 1)

    # ------------------------------------------------------------------------------------------------------------------
    # Fix an annoying bug in the ultimatelistctrl whereby sometimes the list doesn't get redrawn.  This seems to
    # happen with page down and large page jumps.
    def on_page_changed(self, event) -> None:
        wx.CallAfter(self.list.SendSizeEvent)
        self.list.Refresh()
        event.Skip()

    # ------------------------------------------------------------------------------------------------------------------
    # Delete the currently selected file from the favorites list
    def delete_favorite(self) -> None:
        self.video_stop()  # Stop the video
        results = self.get_selected_row_rom_file_name()
        rom_file_name = results['rom_file_name']
        rom_name = results['rom_name']
        system_file_name = results['system_file_name']
        favorites_flag = results['favorites_flag']

        # Only try and remove if it is flagged as a favorite
        if favorites_flag == 1 and rom_file_name != '':
            # If it's a system then we have to reformat the file name to remove as it is structured with the -->
            if system_file_name != '':
                rom_file_name = '-->'.join([rom_file_name, rom_name, system_file_name])

            dlg = wx.MessageDialog(None,
                                   _('Are you sure you want to remove {} from favorites?').format(rom_file_name),
                                   _('Confirm Remove From Favorites'),
                                   wx.YES_NO | wx.NO_DEFAULT | wx.ICON_QUESTION)
            if dlg.ShowModal() == wx.ID_YES:
                # Now we have to remove this ROM name from the favorites.dat file
                try:
                    with open(os.path.join(path_to_me, 'favorites.dat'), 'r') as fp:
                        lines = fp.readlines()
                    with open(os.path.join(path_to_me, 'favorites.dat'), 'w') as fp:
                        for line in lines:
                            if line != ''.join([rom_file_name, '\n']):
                                fp.write(line)
                except FileNotFoundError:
                    msg = _('ERROR:  Unable to find the favorites.dat file in the pfemame folder')
                    self.my_logger.error(msg)
                    self.messages_append(f'{msg}\n', True)
                # Now force the favorites list to be reloaded and the list refreshed to remove the favorites flag

                # Check each ROM in the favorites file and set the favorites flag in mylist
                self.get_dat_file_data('favorites.dat')

                # Refresh the ROM list control now that the game is removed
                self.on_filter_rom_list(1)

                msg = _('Romset {} removed from favorites').format(rom_file_name)
                self.statusbar_panel.set_status(msg)
                self.my_logger.info(msg)
                self.messages_append(f'{msg}\n', True)
                self.write_games_database()  # Write out the ROM database (mylist) to a cache file
            dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # This function gets the currently selected row in the list control.
    # It then looks at various columns and returns their values
    def get_selected_row_rom_file_name(self) -> dict[str]:
        current_item = int(self.list.GetFirstSelected())  # Get the currently selected line in the ROM list control

        if current_item != -1:  # Only try to get the text if we have a row selected
            rom_name = self.list.GetItem(current_item, index_rom_name)
            rom_file_name = self.list.GetItem(current_item, index_rom_file_name)
            rom_type = self.list.GetItem(current_item, index_category)
            system_file_name = self.list.GetItem(current_item, index_system_program_file_name)
            game_rating = self.list.OnGetItemColumnImage(current_item, index_rating)[0] - 2
            short_name = self.list.GetItem(current_item, index_system_short_name)
            favorites_flag = self.list.OnGetItemColumnImage(current_item, index_favorite) - 8

            return {
                "rom_name": rom_name.GetText(),
                "rom_file_name": rom_file_name.GetText(),
                "rom_type": rom_type.GetText(),
                "system_file_name": system_file_name.GetText(),
                "game_rating": game_rating,
                "short_name": short_name.GetText(),
                "favorites_flag": favorites_flag,
            }

        else:
            return {
                "rom_name": '',
                "rom_file_name": '',
                "rom_type": '',
                "system_file_name": '',
                "game_rating": '',
                "short_name": '',
                "favorites_flag": ''
            }

    # ------------------------------------------------------------------------------------------------------------------
    # Function to process the single click ROM selection
    # This updates the snapshot image for the ROM (if it exists) and also the ROM & MAME info windows
    def on_single_click(self, _event) -> None:
        self.mc_forced_stop()  # Stop the game video player
        self.inactivity_timer_start()  # Reset the inactivity timer
        results = self.get_selected_row_rom_file_name()

        rom_file_name = results['rom_file_name']
        system_file_name = results['system_file_name']
        short_name = results['short_name']

        # Check if a video folder paths exists, and if it does, do we have a valid video file for this game.  If so
        # then show (unhide) the load video button
        no_video_text = _('No Video Available')
        if os.path.isdir(self.foldersVideo) and not self.video_disable:
            # Get the video file name if one exists by using a wildcard
            try:
                file_match = glob.glob(os.path.join(self.foldersVideo, f'{rom_file_name}.*'))[0]
            except IndexError:
                file_match = ''
            if self.foldersVideo != '' and os.path.isfile(file_match):
                self.videoLoadCaption.SetLabel(_('Video Available - Click to Play'))
                self.videoLoadCaption.Enable()
                self.video_to_play = file_match
                # If we have our preferences set to autoload the video (and we have the snapshots selected)
                if self.prefVideoAutoload == 'True' and self.image_options_index == 0:
                    # self.videoLoadCaption.Disable()
                    self.videoLoadCaption.SetLabel(_('Auto loading video in 5 seconds'))
                    self.videoTimer.Start(1000)  # Start the 5 second timer to trigger the video autoplay
            else:
                self.videoLoadCaption.Disable()
                self.videoLoadCaption.SetLabel(no_video_text)
                self.video_to_play = ''
        else:
            if self.video_disable:
                no_video_text = _('Video Playback Disabled')
            self.videoLoadCaption.Disable()
            self.videoLoadCaption.SetLabel(no_video_text)
            self.video_to_play = ''

        # Determine what static image should be displayed depending upon the image option selected
        image_path = self.image_options[self.image_options_index][0] if self.image_options else ''

        # Check to see if there is a folder with the image name
        # If there is, grab the first image which should be called 0000.png
        # Otherwise just grab the normal snapshot image
        if os.path.isdir(os.path.join(image_path, rom_file_name)):
            # If it's a system then look for the system program name
            if system_file_name != '':
                # Firstly try to get the short name version of the image file.  If that can't be found, then try and
                # get the long name image file (If a rom is run with a short name, when you take a snapshot it will
                # save the image using the short name, so you really want to use that one).
                index = next(
                    i for i, x in enumerate(self.mylist) if x[index_system_program_file_name] == system_file_name)

                def get_file_name(name_to_check):
                    program_name = self.mylist[index][name_to_check]
                    if rom_file_name != 'visualpinball':
                        # Strip off the . and file name suffix (e.g. .rom)
                        program_name = os.path.splitext(program_name)[0]
                    return os.path.join(image_path, rom_file_name, f'{program_name}.png')

                # Try to get an image file name for short name.  If it fails, try to get one for the long file name.
                file_name = get_file_name(index_system_short_name)
                if not os.path.exists(file_name):
                    file_name = get_file_name(index_system_program_file_name)
            else:
                # Build the file name for the snapshot image using the full path
                file_name = os.path.join(image_path, f'{rom_file_name}/0000.png')
        else:
            # Build the file name for the snapshot image using the full path
            file_name = os.path.join(image_path, f'{rom_file_name}.png')
        self.display_rom_image(file_name)  # Update the ROM image
        
        output = self.get_rom_info(rom_file_name, system_file_name, short_name)  # Get the ROM information
        self.controlRomInfo.SetValue(output)
        self.controlRomInfo.SetInsertionPoint(0)  # Move the scrollbar back to the top of the ROM info windows

        output = self.get_mameinfo_info(rom_file_name, system_file_name, short_name)  # Get the MAME information
        self.controlMAMEInfo.SetValue(output)
        self.controlMAMEInfo.SetInsertionPoint(0)  # Move the scrollbar back to the top of the ROM info window

    # ------------------------------------------------------------------------------------------------------------------
    # This provides a basic override function for the softwarelist HASH file short names.
    # For systems, to get the software HASH file support (needed for cheats), you have to run the system program
    # using the HASH (XML) file short name, not the program file name.  The problem is that some of these do not seem to
    # be working properly (e.g. ZX81 and ZX Spectrum tape files).  To get around this I have added this override feature
    # at least for now, until either the problem is solved, or I find out that I have been doing it wrong (honestly, the
    # documentation in this area of MAME is virtually non existent).
    def on_suppress_short_version(self, _event) -> None:
        self.on_double_click(suppress=False)

    # ------------------------------------------------------------------------------------------------------------------
    # Start a random ROM
    # This will search for a random ROM that has the exists flag set (no point running one that doesn't exist).
    # It won't however be able to determine that the selected ROM is working.
    def on_random(self, _event) -> None:
        _event.Skip()
        rom_file_name = self.get_selected_row_rom_file_name()['rom_file_name']

        # Make sure there is actually a rom file selected (When application is first run there isn't)
        # This just prevents odd errors in the output
        if rom_file_name == '':
            self.my_logger.error(_('Tried running a ROM without a file name selected'))
            return  # Exit the function as there is no point proceeding
        else:
            self.comboBoxFilter.SetValue(_('Exists'))
            self.controlSearch.SetValue('')  # Clear the search field
            self.on_filter_rom_list(-1)  # Force an update of the game list

            # Try to find an existing ROM 200 times then give up
            for i in range(200):
                random_selection = randint(0, self.list.GetItemCount() - 1)
                file_name = self.list.GetItem(random_selection, 1).GetText()
                # Now look for this filename in the mylist and get the exists flag
                # (you can't get it from the displayed list as it's an icon)
                index = next(i for i, x in enumerate(self.mylist) if x[index_rom_file_name] == file_name)
                exists = self.mylist[index][3]
                if exists == '1':
                    self.list.Select(random_selection)
                    self.list.EnsureVisible(random_selection)
                    self.on_filter_rom_list(-1)  # Force an update of the game list
                    wx.GetApp().Yield()
                    self.on_double_click()
                    break
            else:
                # You should never ever end up here as it should only be looking through a list of ROMs flagged as
                # existing but as a just in case I have included this as I haven't yet proven that it can never
                # ever ever happen.
                dlg = wx.MessageDialog(self,
                                       _('I am having trouble finding a ROM to play that exists.  '
                                       'Do you actually have any?'),
                                       _('ERROR'),
                                       wx.OK | wx.ICON_INFORMATION)
                dlg.ShowModal()
                dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # Function to process the double click event to run a game
    #
    # Additional Flags info;
    # suppress - This will prevent the software list HASH file from being used for systems and will just run the
    #            system based on the file name.  This is needed as some systems may not have the right checksum
    #            or some other detail contained in the HASH file and therefore won't run the program.
    # engine - The emulation engine to use when running the rom.  Options are mame, model2, model3, and pinball.
    # ignore_columns - If this is False, then double clicking in any column will run the game.  If its True then it
    #                  will check if the mouse pointer is in the favorite column or the ratings column.  If its in
    #                  one of those, a different action will be performed (changing game rating or adding / removing
    #                  from the favorites list).
    # input_playback_or_record - Tells the function to send info to rungame to that when the MAME rom is executed,
    #                            a command is sent to record or play back previously recorded input file.
    # pb_r_filename - This is the file name for the above
    def on_double_click(self,
                        _event=None,
                        suppress: bool = True,
                        force_mame: bool = False,
                        ignore_columns: bool = False,
                        input_playback_or_record: str = '',
                        pb_r_filename: str = '',
                        rom_file_name: str = '',
                        system_file_name: str = '') -> None:
        if _event is not None:
            _event.Skip()

        self.inactivity_timer.Stop()  # Stop the inactivity timer so we don't trigger screensaver when playing a ROM

        results = self.get_selected_row_rom_file_name()
        # Allow a double click in the favorites column or the game rating column to make a change to those settings
        # but not actually run the game
        if self.start_up_complete is True and ignore_columns is False and self.prefClickFavRating == 'True':
            column_selected = functions.get_column_clicked(self.list)  # Get the column that was clicked in
            if column_selected == index_favorite:
                favorites_flag = results['favorites_flag']
                # If the ROM is not currently flagged as a favorite, then make it one, otherwise remove it from
                # the favorites list and delete the flag
                if favorites_flag == 0:
                    self.on_add_favorite(-1)
                else:
                    self.delete_favorite()
                return
            if column_selected == index_rating:
                self.on_set_rating(-1)
                return

        self.video_stop()  # Stop the video

        if rom_file_name == '':
            rom_file_name = results['rom_file_name']
            rom_name = results['rom_name']
            system_file_name = results['system_file_name']
        else:
            rom_name = ''

        # Get the system folder name based upon what has already been stored in mylist (assuming one is needed)
        try:
            index = next(i for i, x in enumerate(self.mylist) if x[index_system_program_file_name] == system_file_name)
            system_folder_name = self.mylist[index][index_system_folder_name]
        except (StopIteration, IndexError):
            system_folder_name = ''

        short_name = results['short_name']

        #  Check if we are suppressing the softwarelist HASH file short name version for system files
        if suppress and self.pref_short_name_global_suppress != 'True':
            # short_name = results['short_name']
            suppress_short_name = False
        else:
            # short_name = ''
            suppress_short_name = True

        # Make sure there is actually a rom file selected (When application is first run there isn't)
        # This just prevents odd errors in the output
        if rom_file_name == '':
            self.my_logger.error(_('Tried running a ROM without a file name selected'))
            return  # Exit the function as there is no point proceeding

        if system_file_name != '':
            msg = _('System {} started with program {} selected').format(rom_name, system_file_name)
        else:
            msg = _('Romset {} started').format(rom_name)
        self.my_logger.info(msg)  # Write to the log file
        self.messages_append(f'{msg}\n')

        # Get the folder the mame.ini file resides in (Should be the same as the other MAME folders
        folder_mame_ini = os.path.dirname(self.path_to_mame_ini_file)

        self.joystick_block = True  # Block the joystick events from being triggered when running a ROM
        self.inactivity_timer_block = True  # Block the inactivity timer from being started while running a ROM

        # Call the rungame class (imported file) and make this party happen.  This is done in a separate thread.
        rungame.RunGame(self,
                        rom_file_name,
                        self.path_mame_exe,
                        'Normal View',
                        rom_name,
                        system_file_name,
                        short_name,
                        system_folder_name,
                        folder_mame_ini,
                        self.folder_mapping,
                        suppress_short_name,
                        input_playback_or_record,
                        pb_r_filename,
                        force_mame
                        )

    # ------------------------------------------------------------------------------------------------------------------
    # Process the thread request to update the main process gui.
    # This one is to send data to the messages control
    def update_display_messages(self, msg: str) -> None:
        self.messages_append(msg)

    # ------------------------------------------------------------------------------------------------------------------
    # Process the thread request to update the main process gui.
    # This one is to send data to the logfile
    def update_logfile(self, msg: str) -> None:
        self.my_logger.info(msg)  # Write to the log file

    # ------------------------------------------------------------------------------------------------------------------
    # Process the thread request to update the main process gui.
    # This one is to send data pop-up messages
    def update_display_popups(self, msg: str) -> None:
        dlg = wx.MessageDialog(None,
                               msg,
                               _('Important Notification'),
                               wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # Process the thread request to update the main process gui.
    # This one is to display a statusbar message
    def update_status(self, msg: str) -> None:
        self.statusbar_panel.set_status(msg)

    # ------------------------------------------------------------------------------------------------------------------
    # Process the thread request to update the main process gui.
    # Update the number of runs for the currently selected game
    def update_display_number_of_runs(self, attr1: str, attr2: str, attr3: str, attr4: bool, attr5: str) -> None:
        self.update_number_runs(attr1, attr2, attr3, attr4, attr5)  # Update the number of runs for the ROM

    # ------------------------------------------------------------------------------------------------------------------
    # Processes the CTRL-F keypress capture and set up, so you can just start typing to find a ROM
    def on_find(self, _event) -> None:
        self.controlSearch.SetFocus()  # Set focus to the search input control
        self.controlSearch.SetSelection(-1, -1)  # Select all text in the control

    # ------------------------------------------------------------------------------------------------------------------
    # Some cleanup required for the on_close and on_restart events
    def cleanup(self) -> None:
        self.close_selected = True  # Get the VLC playback while loop to exit
        time.sleep(1)  # Give the while loop enough time to exit

        # Store the currently selected view and row so that they can be recalled on startup
        view = self.comboBoxFilter.GetValue()
        results = self.get_selected_row_rom_file_name()
        rom_name = results['rom_name']
        system_file_name = results['system_file_name']
        if rom_name != '':
            iniFile.set_ini_file_variables([('system', 'lastromselected', rom_name)])
            iniFile.set_ini_file_variables([('system', 'lastsystemselected', system_file_name)])
        iniFile.set_ini_file_variables([('system', 'lastviewselected', view)])

        # Stop joystick related timer events
        self.joystick_block = True  # Stop the polling timer from trying to do anything
        if self.joystick is not None:
            self.joystick_polling_timer.Stop()
            self.joystick.Destroy()  # Delete the joystick instance to kill its thread

        # Stop all other timer related events
        if self.media_player_available:
            self.video_stop()  # Stop the video
            self.videoTimer.Stop()
        self.alert_timer.Stop()
        self.inactivity_timer.Stop()

    # ------------------------------------------------------------------------------------------------------------------
    # Handle the close event and ensure the parent window is given back control
    def on_close(self, _event) -> None:
        self.cleanup()  # Perform some cleanup before closing
        elapsed_time = str(datetime.timedelta(seconds=time.time() - self.startup_process_start_time))
        self.my_logger.info(_('Application Ended after {}').format(elapsed_time))  # Write to the log file
        self.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # Force the pfeMAME application to restart.
    # This allows us to use newly changed variables or settings that need a restart to be performed.
    def on_force_restart(self, msg: str = '') -> None:
        self.on_restart(-1)

    def on_restart(self, _event) -> None:
        self.cleanup()  # Perform some cleanup before restarting

        # Add a flag so that when restarted the application knows that it was forced to restart
        sys_argv = sys.argv
        sys_argv.append('-x')

        python = sys.executable  # Get the full path to the python executable the application is using

        # Check to see if we can actually force a restart of the application.  If we're on macOS we can't do this
        # for now as I have not found a reliable way to work this out (it uses links which don't work).  I've also
        # added a check to make sure we're not looking at a link just in case on other systems.
        if self.system == 'mac' or os.path.islink(python):
            dlg = wx.MessageDialog(self,
                                   _('Please restart the application manually'),
                                   _('! IMPORTANT !'),
                                   wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
            self.on_close(-1)

        else:
            # The following is my WIP attempt at finding the real Python3 path for macOS builds.  It's not currently
            # working but may be a step in the right direction.
            ##########################
            # dirtmp = os.path.dirname(python)
            # dirtemp2 = dirtmp.strip('MacOS')

            # temp = 'Frameworks/Python3.framework/Versions/3.8'
            # newpath = os.path.join(dirtemp2, temp, 'Python3')
            # messageme(newpath)
            # if os.path.isfile(newpath):
            #     messageme('its a file')
            # else:
            #     messageme('error no file')

            # python=newpath
            #############################################

            os.execl(python, python, *sys_argv)

    # ------------------------------------------------------------------------------------------------------------------
    # This function updates the number of runs info for the ROM just run
    def update_number_runs(self,
                           rom_file_name: str,
                           rom_name: str,
                           system_file_name_passed: str,
                           is_a_system: bool,
                           caller: str) -> None:
        results = self.get_selected_row_rom_file_name()
        if rom_name == '':
            rom_name = results['rom_name']

        if rom_file_name == '':
            rom_file_name = results['rom_file_name']

        if caller == 'normal_view':
            system_file_name = results['system_file_name']
        else:
            # The caller must be tile view
            system_file_name = system_file_name_passed if is_a_system is True else ''

        self.my_logger.info(_('Updating number of runs info for ROM {}').format(rom_file_name))  # Write to the log file

        # Increment the number of runs in the games list
        rom_file_name, index = self.get_rom_file_name(system_file_name, rom_name, rom_file_name)
        self.mylist[index][index_number_runs] = str(int(self.mylist[index][index_number_runs]) + 1)

        self.clearSearch = False  # Don't clear the search field when refreshing the games list
        self.on_filter_rom_list(1)  # Force an update of the game list to show the number of runs changed

        # Get all ROMs with runs from the ROMs list and store them in the runs.dat file
        try:
            with open(os.path.join(path_to_me, 'runs.dat'), 'w') as fp:
                for line in self.mylist:
                    if line[index_number_runs] != '0' and line[index_number_runs] != '':
                        # Format it differently if it's a system file
                        if line[index_system_program_file_name] == '':
                            rom_file_name = line[index_rom_file_name]
                        else:
                            rom_file_name = '-->'.join([line[index_rom_name], line[index_rom_file_name],
                                                        line[index_system_program_file_name]])
                        fp.write('='.join([rom_file_name, line[index_number_runs] + '\n']))
        except IOError:  # Issue writing so just ignore
            self.my_logger.error(_('ERROR: Unable to write to runs.dat file'))  # Write to the log file

        self.write_games_database()  # Write out the ROM database (mylist) to a cache file

    # ------------------------------------------------------------------------------------------------------------------
    # This function is the event service for the add rating event
    # It pops up a box requesting a rating between 1 and 5 for the
    # currently selected ROM and then updates the ROM info
    # and refreshes the display.
    def on_set_rating(self, _event) -> None:
        results = self.get_selected_row_rom_file_name()
        rom_file_name = results['rom_file_name']
        rom_name = results['rom_name']
        game_rating = results['game_rating']
        system_file_name = results['system_file_name']
        selected_rom_position = self.list.GetFocusedItem()  # Used to place the selected ROM to the top of the view
        if rom_file_name != '':
            result = romRating.RomRating(rating=int(game_rating)).ShowModal()
            if -1 < result < 6:
                result = str(result)

                rom_file_name, index = self.get_rom_file_name(system_file_name, rom_name, rom_file_name)
                self.mylist[index][index_rating] = result

                self.clearSearch = False  # Don't clear the search field when refreshing the ROM list
                self.on_filter_rom_list(1)  # Force an update of the ROM list to show the ROM rating changed
                self.statusbar_panel.set_status(f'Romset {rom_file_name} rated as {result} stars')

                self.list.EnsureVisible(selected_rom_position)  # Recover the scroll position to this ROM

                # Now we read in the current ratings.dat file and strip out the selected ROM if it exists
                ratings = []
                try:
                    with open(os.path.join(path_to_me, 'ratings.dat'), 'r') as fp:
                        ratings = [line for line in fp if line.split('=')[0] != rom_file_name]
                except IOError:  # File doesn't exist so just ignore
                    self.my_logger.warning(_('WARNING: Failed to read ratings.dat file'))  # Write to the log file

                # Add the currently selected ROM and its new rating
                ratings.append('='.join([rom_file_name, result + '\n']))

                # And now we write it back out again with the updated ROM ratings
                try:
                    with open(os.path.join(path_to_me, 'ratings.dat'), 'w') as fp:
                        fp.writelines(ratings)
                except IOError:  # Issue writing so just ignore
                    self.my_logger.error(_('ERROR: Unable to write to ratings.dat file'))  # Write to the log file

                self.my_logger.info(_('Romset {} rating updated').format(rom_file_name))
                self.messages_append(_('Romset {} rating updated\n').format(rom_file_name), True)
                self.write_games_database()  # Write out the ROM database (mylist) to a cache file

    # ------------------------------------------------------------------------------------------------------------------
    # This function takes the currently selected ROM and adds it to the favorites file
    # A check is performed before adding to the favorites file to ensure it doesn't already
    # exist (to prevent duplicates)
    def on_add_favorite(self, _event) -> None:
        results = self.get_selected_row_rom_file_name()
        rom_file_name = results['rom_file_name']
        rom_name = results['rom_name']
        system_file_name = results['system_file_name']
        if rom_file_name != '':
            # Save the currently selected ROM position so it can be recalled later
            selected_rom_position = self.list.GetFocusedItem()
            self.my_logger.info(_('Adding ROM {} to favorites').format(rom_file_name))  # Write to the log file

            rom_file_name, index = self.get_rom_file_name(system_file_name, rom_name, rom_file_name)
            self.mylist[index][index_favorite] = '1'  # Flag this ROM as a favorite

            # Read from the favorites file and store in a temporary list
            temp_list = []
            try:
                with open(os.path.join(path_to_me, 'favorites.dat'), 'r') as fp:
                    temp_list = fp.readlines()
            except IOError:  # The favorites file doesn't currently exist so no problems, just ignore the error
                self.my_logger.warning(_('WARNING: Unable to read from favorites.dat file'))  # Write to the log file

            # Now add the new favorite ROM to the list
            temp_list.append(''.join([rom_file_name, '\n']))
            new_set = set(temp_list)  # Prevent any duplicates by converting to a set

            # Write out the new favorites file
            try:
                with open(os.path.join(path_to_me, 'favorites.dat'), 'w') as fp:
                    fp.writelines(new_set)
            except IOError:  # Issue writing so just ignore
                self.my_logger.error(_('ERROR: Unable to write to favorites.dat file'))  # Write to the log file

            self.clearSearch = False  # Don't clear the search field when refreshing the ROM list
            self.on_filter_rom_list(1)  # Force an update of the ROM list to show the game now selected as a favorite
            msg = _('Romset {} added to favorites').format(rom_file_name)
            self.statusbar_panel.set_status(msg)
            self.my_logger.info(msg)
            self.messages_append(f'{msg}\n', True)

            self.write_games_database()  # Write out the ROM database (mylist) to a cache file
            self.list.EnsureVisible(selected_rom_position)  # Recover the scroll position to this ROM

    # ------------------------------------------------------------------------------------------------------------------
    # Get the rom_file_name and index for the match required
    def get_rom_file_name(self, system_file_name: str, rom_name: str, rom_file_name: str) -> (str, int):
        if system_file_name == '':
            index = next(i for i, x in enumerate(self.mylist) if x[index_rom_name] == rom_name)
        else:
            index = next(i for i, x in enumerate(self.mylist) if (x[index_system_program_file_name] == system_file_name
                                                                  and x[index_rom_file_name] == rom_file_name))
            rom_file_name = '-->'.join([rom_file_name, rom_name, system_file_name])
        return rom_file_name, index

    # ------------------------------------------------------------------------------------------------------------------
    # Get the requested DAT file data
    #
    # This is used for runs, favorites, and ratings.DAT files
    def get_dat_file_data(self, file_to_update: str) -> None:
        self.my_logger.info(_('Reading ROM {} flags').format(os.path.splitext(file_to_update)[0]))  # Write to the log file
        self.statusbar_panel.set_status(_('Reading ROM {} flags.  Please Wait.........').format(os.path.splitext(file_to_update)[0]))
        if self.mylist:
            # If it's the favorites.dat file then set all favorites flags to zero.
            # This is required in order to handle a DELETE operation.
            if file_to_update == 'favorites.dat':
                for row in self.mylist:
                    row[index_favorite] = '0'
            error_count = 0
            try:
                # Now try to read out the data in the dat file
                with open(os.path.join(path_to_me, file_to_update), 'r') as fp:
                    for line in fp:
                        if file_to_update == 'favorites.dat':
                            rom_file_name = line.strip()  # Remove the linefeed
                        else:
                            line = line.split('=')
                            if len(line) == 2:  # Check that the line is formatted correctly
                                rom_file_name = line[0]
                                data = line[1].rstrip()
                            else:
                                break

                        # Check to see if it's a system file name - if so strip out the garbage
                        rom_file_name_split = rom_file_name.split('-->')
                        if len(rom_file_name_split) > 2:
                            # This is a system file with a system file name to run
                            index = functions.match_list_location(self.mylist, 2, rom_file_name_split[2])
                        elif len(rom_file_name.split('-->')) > 1:
                            # This is a top level system file only - no system file name to run
                            index = functions.match_list_location(self.mylist, 1, rom_file_name_split[1])
                        else:
                            # Not a system file, just a normal ROM
                            index = functions.match_list_location(self.mylist, index_rom_file_name, rom_file_name)

                        if index is not None:
                            if file_to_update == 'favorites.dat':
                                self.mylist[index][index_favorite] = '1'
                            elif file_to_update == 'runs.dat':
                                self.mylist[index][index_number_runs] = data
                            else:
                                self.mylist[index][index_rating] = data
                        else:
                            error_count += 1
                            self.my_logger.warning(
                                _('A file name called {} in the {} file does not exist in the '
                                  'MAME ROM list').format(rom_file_name, file_to_update))  # Write to the log file

            except IOError:  # DAT file doesn't exist so just ignore this part
                self.my_logger.warning(_('WARNING: {} file does not exist').format(file_to_update))  # Write to the log file

            if error_count > 0:
                self.messages_append(_('{} error(s) found in {} file.  See logfile.log for '
                                       'more information.\n').format(error_count, file_to_update), True)

            self.my_logger.info(_('Completed reading ROM {} flags').format(os.path.splitext(file_to_update)[0]))

    # ------------------------------------------------------------------------------------------------------------------
    # This function reads the categories.dat file and stores it in a list
    def load_categories(self) -> None:
        try:
            with open(os.path.join(path_to_me, 'categories.dat'), 'r') as fp:
                self.categories = fp.read().splitlines()
        except IOError:
            self.categories = []
            self.my_logger.error(_('Unable to locate categories.dat file.  '
                            'Select some ROM categories by checking the box for the category.'))

    # ------------------------------------------------------------------------------------------------------------------
    # This function creates the filtered game list for display.
    # To use the filter, pass the mylist column and value you wish to filter for.
    # To just show all, pass a -1 as the column and empty string as value
    def populate_list(self, filtercolumn: int, filtervalue: str, read_from_cache: int) -> int:
        # If this is true we look at the previously used filter value and use those to
        # define how to filter.  This is useful to prevent reverting to an unwanted filter view
        # after some form of refresh operation
        if read_from_cache is True:
            filtercolumn = self.filtercolumn_cache
            filtervalue = self.filtervalue_cache

        self.filtercolumn_cache = filtercolumn
        self.filtervalue_cache = filtervalue

        MainFrame.mylist_filtered = []
        append = MainFrame.mylist_filtered.append
        lower = str.lower
        filtervalue_split = filtervalue.split()  # Turn the search string into a list of strings by each word
        joins = '\n'.join
        categories_local = self.categories

        for line in self.mylist:
            # Check the filter requirements
            # Then check if the game category is listed in the categories.dat file
            # If the categories.dat file doesn't exist, or the game category is blank,
            # then display it anyway just in case.
            #
            # If the filter is for the game name (type ahead search) then also allow a match with other fields 
            # as follows below.
            # They are checked in the order of;
            #   rom file name
            #   rom name
            #   system name
            #   category
            #   year
            #   manufacturer

            if filtercolumn == 1:
                # This is for the dynamic search
                filtercolumnmatch = False
                # Build the string of space-separated values from the mylist line currently being looked at.  This is
                # the source data that will then be searched for a match.  If there is a match then this line is
                # added to the new list.
                test_line = lower(joins(
                    [line[index_rom_name], line[index_rom_file_name], line[index_system_program_file_name],
                     line[index_category], line[index_year], line[index_manufacturer]]))

                # Only display matches for ROMs that have a category selected in the categories list
                if ([e for e in filtervalue_split if e in test_line] == filtervalue_split) and (
                        line[index_category] in categories_local):
                    filtercolumnmatch = True
            else:
                # Otherwise just a match with the selected column, e.g. favorites or exists flags
                filtercolumnmatch = True if filtervalue in lower(line[filtercolumn]) else False

            if (filtercolumnmatch is True or (filtercolumn == -1)) and line[index_category] in self.categories:
                #  Only display matches for ROMs that have a category selected in the categories list.
                #  That way I don't show games I am not interested in.
                append(line)

        self.list.SetItemCount(len(MainFrame.mylist_filtered))  # Update the virtual list control with the data length
        self.list.Refresh()  # Refresh the control
        return len(MainFrame.mylist_filtered)

    # ------------------------------------------------------------------------------------------------------------------
    # This function processes the game filter combo box selection event
    def on_filter_rom_list(self, event) -> None:
        # if self.list_notebook.GetSelection() != 0:
        #     self.list_notebook.SetSelection(0)  # Make sure the main list view is shown
        #     self.list_notebook_current = 0

        read_from_cache = True if event == 1 else False

        match self.comboBoxFilter.GetValue():
            case 'All':  # All
                filtercolumn = -1
                filtervalue = ''
            case 'Favorites':  # Favorites
                filtercolumn = index_favorite
                filtervalue = '1'
            case 'Exists':  # Exists
                filtercolumn = index_exists
                filtervalue = '1'
            case _:   # Else show missing roms
                filtercolumn = index_exists
                filtervalue = '0'

        number_roms_in_filtered_list = self.populate_list(filtercolumn=filtercolumn,
                                                          filtervalue=filtervalue,
                                                          read_from_cache=read_from_cache)
        self.update_rom_numbers(number_roms_in_filtered_list)

        try:
            if self.clearSearch is True:
                self.controlSearch.Enable(False)
                self.clear_event_override = True  # Prevent the search input event from triggering
                self.controlSearch.SetValue('')  # Empty the search input box
                self.controlSearch.Enable()
            # Do this so that if the evtFilterGamesList event is triggered by the drop-down event,
            # the search box is reset
            self.clearSearch = True
        except NameError:
            self.clearSearch = True

        self.clear_event_override = False  # Allow the search input event to be triggered now

    # ------------------------------------------------------------------------------------------------------------------
    # This function stops the video player and displays the right caption depending upon if there was a video to
    # play, or not.
    def video_stop(self) -> None:
        if not self.video_disable:
            self.videoTimer.Stop()  # Make sure the video timer is stopped
            self.mc_forced_stop()  # Stop the ROM video player
            text = _('No Video Available') if self.video_to_play == '' else _('Video Stopped')
            self.videoLoadCaption.SetLabel(text)

    # ------------------------------------------------------------------------------------------------------------------
    # This function processes the search box event trigger to allow for dynamic searching
    def on_search_dynamic(self, event) -> None:

        # Don't allow the event to be processed if we have suppressed the event by setting the following flag.  This is
        # done to prevent the selection of the views drop down box causing the search control to be cleared and thus
        # triggering the search event when then resets the view drop down box (a cyclic issue).
        if self.clear_event_override is False:
            self.video_stop()  # Stop the video
            search_text = self.controlSearch.GetValue().lower()  # Get the search text converted to lower case
            self.clearSearch = False
            self.comboBoxFilter.SetValue(_('All'))
            self.comboBoxFilter.Refresh()
            self.clearSearch = True

            number_roms_in_filtered_list = self.populate_list(filtercolumn=1,
                                                              filtervalue=search_text,
                                                              read_from_cache=False)

            self.update_rom_numbers(number_roms_in_filtered_list)

            if number_roms_in_filtered_list > 0:
                self.list.EnsureVisible(0)  # Force the first row to be visible
            event.Skip(True)

    # ------------------------------------------------------------------------------------------------------------------
    # Update the rom quantity counters in the status bar
    def update_rom_numbers(self, number_roms_in_filtered_list: int) -> None:
        self.statusbar_panel.set_quantities(f'{str(len(self.mylist))} ',
                                            f'{str(self.availableRoms)} ',
                                            f'{str(number_roms_in_filtered_list)}')

    # ------------------------------------------------------------------------------------------------------------------
    # This function updates the ROM image placeholder
    # Images are scaled to size of 224 x 256 pixels (x,y)
    def display_rom_image(self, file_name: str) -> None:
        # Load the image or if none found then set default
        _nullTarget = wx.LogNull()  # Prevent a crash with malformed png files in windowsXP
        self.Freeze()
        try:
            with open(file_name):
                # Load the image and scale it to best fit while maintaining its aspect ratio
                img = wx.Image(file_name, wx.BITMAP_TYPE_PNG)
                image_width, image_height = img.GetSize()
                if image_width > image_height:
                    scale_x = self.image_size_x
                    scale_y = int(image_height * (scale_x / image_width))
                elif image_height > image_width:
                    scale_x = int(image_width * (self.image_size_y / image_height))
                    scale_y = self.image_size_y
                else:
                    scale_x = self.image_size_x
                    scale_y = self.image_size_y
                img = img.Scale(scale_x, scale_y)
        except IOError:
            img = wx.Image(default_image_file_name, wx.BITMAP_TYPE_ANY).Rescale(self.image_size_x, self.image_size_y)
        except wx._core.PyAssertionError:
            # I have done this to capture a core error when the rescale function tries to rescale an image that isn't
            # quite right.  There is still an annoying pause while the function tries to resize but at least it
            # doesn't throw an error
            img = wx.Image(default_image_file_name, wx.BITMAP_TYPE_ANY).Rescale(self.image_size_x, self.image_size_y)
            msg = _('The snapshot image {} '
                    'is causing an error when trying to resize it.  It may be corrupted.').format(file_name)
            self.my_logger.error(msg)  # Write to the log file
            self.messages_append(f'ERROR: {msg}\n')
            self.enable_alert_timer()

        del _nullTarget
        self.Thaw()
        img = img.ConvertToBitmap()

        # Draw the magnifier in the bottom RHS of the image
        x, y = img.GetSize()
        dc = wx.MemoryDC(img)
        x = x - self.overlay_image_size[0]
        y = y - self.overlay_image_size[1]
        dc.DrawBitmap(self.overlay_image, x, y, True)
        del dc
        self.image_panel.image.SetBitmap(
            wx.Bitmap(img))  # Convert it to a wx.Bitmap, and put it on the wx.StaticBitmap
        self.image_panel.Layout()  # Force the centering of the image

        # Store the image file name so that we can access it from the popupimage viewer
        MainFrame.image_file_name = file_name

    # ------------------------------------------------------------------------------------------------------------------
    # Open the File Paths child window
    def on_file_paths(self, _event) -> None:
        filepaths.WindowFilePaths(self.my_logger)

    # ------------------------------------------------------------------------------------------------------------------
    # Execute the mame command '-verifyroms' and pass the output to the message box.  Then do the same using
    # -verifysamples but only pass the output if there are supposed to be samples.
    def on_verify_roms(self, _event) -> None:
        rom_file_name = self.get_selected_row_rom_file_name()['rom_file_name']
        if rom_file_name != '':
            self.messages_append(_('Verifying ROM image for {}\n').format(rom_file_name))

            # Build up the command for -verifyroms
            cmd = [self.path_mame_exe, '-rompath', self.folders_roms, '-verifyroms', rom_file_name]
            p = subprocess.run(cmd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               shell=self.shell_to_use,
                               universal_newlines=True)
            out = p.stdout
            err = p.stderr

            if out != '':
                self.messages_append(out)
            if err != '':
                self.messages_append(err)
            self.controlMessages.AppendText(messageDelimiter)

            # Preserve the existing path, and change paths to the MAME executable
            oldpath = self.path_to_me
            if self.system != 'linux':
                os.chdir(os.path.dirname(self.path_mame_exe))

            # Build up the command for -verifysamples
            self.messages_append(_('Verifying sample files for {}\n').format(rom_file_name))
            cmd = [self.path_mame_exe, '-verifysamples', rom_file_name]
            p = subprocess.run(cmd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               shell=self.shell_to_use,
                               universal_newlines=True)
            out = p.stdout
            err = p.stderr
            os.chdir(oldpath)  # Restore the current working directory

            # Only generate the output message if the romset is supposed to have a sampleset
            if (out == '' or out == 'not required!') and err == '':
                out = _('No Sample Files Required\n')
            if err != '':
                out = '\n'.join([out, err])
            self.messages_append(out, True)

            if self.folders_samples == '' or self.folders_samples == '.':
                msg = _('WARNING:  The MAME samples folder has not been set in filepaths\n')
                self.messages_append(msg)

    # ------------------------------------------------------------------------------------------------------------------
    # This function executes the MAME command -listmedia for the selected rom.
    # It is intended to be used with systems (e.g. ZX Spectrum) where the system is capable of running a number
    # of types of program with different media extensions (e.g. .tap, .bin)
    def on_view_media_types(self, _event) -> None:
        rom_file_name = self.get_selected_row_rom_file_name()['rom_file_name']
        if rom_file_name != '':
            oldpath = self.path_to_me  # Need to save the current working directory so it can be restored later
            if self.system != 'linux':
                os.chdir(os.path.dirname(self.path_mame_exe))
            extensions = functions.get_system_media_types(self.path_mame_exe, rom_file_name, self.shell_to_use)
            os.chdir(oldpath)

            # Display the result in the messages window
            self.messages_append(_('Media types associated with ROM {}:\n').format(rom_file_name))
            for block in extensions:
                self.controlMessages.AppendText(f'{block} ')
            self.controlMessages.AppendText('\n')
            self.messages_append(messageDelimiter)

    # ------------------------------------------------------------------------------------------------------------------
    # Get MAME rom information
    def on_rom_info(self, _event) -> None:
        results = self.get_selected_row_rom_file_name()
        rom_file_name = results['rom_file_name']
        rom_name = results['rom_name']
        if rom_file_name != '':
            rominfo.WindowMessage(rom_file_name=rom_file_name, rom_name=rom_name, path_executable=self.path_mame_exe)

    # ------------------------------------------------------------------------------------------------------------------
    # This function searches the history file for the ROM information.
    # It supports the older DAT file format and the newer XML file format.
    def get_rom_info(self, rom_file_name: str, system_file_name: str, short_name: str) -> str:
        output = functions.get_rom_info(self.pathHistory, 
                                        self.history_data, 
                                        self.mylist, 
                                        rom_file_name, 
                                        system_file_name, 
                                        short_name, 
                                        index_system_folder_name, 
                                        'history')
        return output

    # ------------------------------------------------------------------------------------------------------------------
    # This function searches the mameinfo.dat file for the ROM information.
    def get_mameinfo_info(self, rom_file_name: str, system_file_name: str, short_name: str) -> str:
        output = functions.get_rom_info(self.pathMAMEInfoDat, 
                                        self.mameinfo_data, 
                                        self.mylist, 
                                        rom_file_name, 
                                        system_file_name, 
                                        short_name, 
                                        index_system_folder_name, 
                                        'mameinfo')
        return output

    # ------------------------------------------------------------------------------------------------------------------
    # This function passes the application version number to the 'about' dialog box
    # and shows the 'about box'
    @staticmethod
    def on_about_box_display(_event) -> None:
        aboutbox.OnAboutBox(version=__version__, copy_right=__copyright__)

    # ------------------------------------------------------------------------------------------------------------------
    # This function displays the MAME in-game controls window
    @staticmethod
    def on_controls_box_display(_event) -> None:
        scrolledwindow.DisplayScrolledWindow(caption=_('MAME Controls'),
                                             file_path=os.path.join(path_to_me, 'controls.txt'),
                                             scroll=False,
                                             size=(500, 600))

    # ------------------------------------------------------------------------------------------------------------------
    # This function displays the pfeMAME logfile in a window
    @staticmethod
    def on_logfile_display(_event) -> None:
        scrolledwindow.DisplayScrolledWindow(caption=_('pfeMAME Log File'),
                                             file_path=os.path.join(path_to_me, 'logfile.log'),
                                             scroll=True,
                                             size=(800, 500))

    # ------------------------------------------------------------------------------------------------------------------
    # This function displays the pfeMAME changelog file in a window
    @staticmethod
    def on_changelog_display(_event) -> None:
        try:
            doc = docx.Document(os.path.join(path_to_me, 'changelog.docx'))  # Creating word reader object.
            full_text = [para.text for para in doc.paragraphs]
            data = '\n'.join(full_text)

            scrolledwindow.DisplayScrolledWindow(caption=_('pfeMAME Changelog File'),
                                                 file_path='',
                                                 scroll=False,
                                                 size=(800, 500),
                                                 text_to_display=data)
        except docx.opc.exceptions.PackageNotFoundError:
            wx.MessageBox(
                _('Unable to find the changelog.docx file.  '
                'Please check that it exists in the pfeMAME folder.'),
                _('ERROR'), wx.OK | wx.ICON_ERROR)

    # ------------------------------------------------------------------------------------------------------------------
    # This function displays the pfeMAME manual
    def on_manual_display(self, _event) -> None:
        manual_path = os.path.join(path_to_me, 'pfeMAME_Manual.pdf')
        if os.path.isfile(manual_path):
            # The Linux / Windows / macOS methods for opening an external PDF are different
            if os.name == 'nt':
                subprocess.run([manual_path], shell=True)
            elif os.name == 'posix':
                types = ['gio',
                         'xdg-open',
                         'gnome-open',
                         'kde-open',
                         'open']
                for i in types:
                    if os.path.exists(f'/usr/bin/{i}'):  # Try the universal Linux version to open the file
                        if self.system == 'mac':
                            subprocess.run([f"/usr/bin/{i} '{manual_path}'"],
                                           shell=True)
                        else:
                            subprocess.run([f"/usr/bin/{i} open '{manual_path}'"],
                                           shell=True)
                        break
                else:
                    # No match found so cant work out how to open the manual
                    wx.MessageBox(
                        _('Sorry but I am unable to find a way to open the pfeMAME manual.  '
                        'Please open the file manually.'),
                        _('ERROR'), wx.OK | wx.ICON_ERROR)
        else:
            wx.MessageBox(
                _('Unable to find the pfeMAME_Manual.pdf file!  '
                'Please check that it exists in the pfeMAME folder.'),
                _('ERROR'), wx.OK | wx.ICON_ERROR)

    # ------------------------------------------------------------------------------------------------------------------
    # This function opens the checkall window to allow all 'existing'
    # ROMs to be checked for correctness
    def on_check_all(self, _event) -> None:
        rom_file_name = self.get_selected_row_rom_file_name()['rom_file_name']
        if rom_file_name != '':
            checkall.CheckAll(path_executable=self.path_mame_exe, folders_roms=self.folders_roms)

    # ------------------------------------------------------------------------------------------------------------------
    # Run the selected ROM and record the inputs for playback later
    def on_input_record(self, _event) -> None:
        if self.folderInput == '':
            wx.MessageBox(
                _('Please set a valid Inputs folder under the File Paths'),
                _('ERROR'), wx.OK | wx.ICON_ERROR)
        else:
            if (rom_file_name := self.get_selected_row_rom_file_name()['rom_file_name']) != '':
                dlg = wx.FileDialog(
                    self,
                    message=_('Create the file for recording inputs for {}').format(rom_file_name),
                    defaultDir=self.folderInput,
                    defaultFile=f'{rom_file_name}.inp',
                    wildcard='INP Files (*.inp)|*.inp',
                    style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
                )
                if dlg.ShowModal() == wx.ID_OK:
                    path = os.path.basename(dlg.GetPath())
                    msg = _('Recording playback for the rom using {}').format(path)
                    self.my_logger.info(msg)  # Write to the log file
                    self.messages_append(f'{msg}\n')
                    self.on_double_click(input_playback_or_record='-record', pb_r_filename=path)

    # ------------------------------------------------------------------------------------------------------------------
    # Run the selected ROM using previously recorded inputs
    def on_input_playback(self, _event) -> None:
        if self.folderInput == '':
            wx.MessageBox(
                _('Please set a valid Inputs folder under the File Paths'),
                _('ERROR'), wx.OK | wx.ICON_ERROR)
        else:
            if (rom_file_name := self.get_selected_row_rom_file_name()['rom_file_name']) != '':
                dlg = wx.FileDialog(
                    self,
                    message=_('Select the file to play back inputs for {}').format(rom_file_name),
                    defaultDir=self.folderInput,
                    defaultFile=f'{rom_file_name}.inp',
                    wildcard='INP Files (*.inp)|*.inp',
                    style=wx.FD_OPEN
                )
                if dlg.ShowModal() == wx.ID_OK:
                    path = os.path.basename(dlg.GetPath())
                    msg = _('Playing the rom using recorded input from {}').format(path)
                    self.my_logger.info(msg)  # Write to the log file
                    self.messages_append(f'{msg}\n')
                    self.on_double_click(input_playback_or_record='-playback', pb_r_filename=path)

    # ------------------------------------------------------------------------------------------------------------------
    # This function allows you to back up the preferences and dat files including the following;
    #
    # categories.dat
    # favorites.dat
    # ratings.dat
    # runs.dat
    # pfemame.ini
    # non_mame_emulators.xml
    def on_backup(self, _event) -> None:

        dlg = wx.DirDialog(self,
                           message=_('Choose a directory for backup:'),
                           defaultPath=path_to_me,
                           style=wx.DD_DEFAULT_STYLE | wx.DD_NEW_DIR_BUTTON
                           )
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            try:
                files_copied = ''.join([functions.copy_one_file(path_to_me, path, file) for file in files_to_copy])

                msg = _('Application preferences and DAT files backed up to {}').format(path)
                self.messages_append(f'{msg}.\n{files_copied}', True)
                self.my_logger.info(msg)  # Write to the log file
            except (TypeError, IOError):
                dlg2 = wx.MessageDialog(self,
                                        _('Unable to copy the application files to the selected folder.  '
                                        'Check the folder permissions.'),
                                        _('ERROR'),
                                        wx.OK | wx.ICON_INFORMATION)
                dlg2.ShowModal()
                dlg2.Destroy()
        else:
            dlg2 = wx.MessageDialog(self,
                                    _('Backup process canceled'),
                                    _('INFORMATION'),
                                    wx.OK | wx.ICON_INFORMATION)
            dlg2.ShowModal()
            dlg2.Destroy()
        dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # This function allows you to restore previously backed up files.  You have to
    # select the folder that was used for the previous backup
    def on_restore(self, _event) -> None:

        dlg = wx.DirDialog(self,
                           message=_('Choose the backup directory to restore application preferences and DAT files from:'),
                           defaultPath=path_to_me,
                           style=wx.DD_DEFAULT_STYLE | wx.DD_NEW_DIR_BUTTON
                           )
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            try:
                files_copied = ''.join([functions.copy_one_file(path, path_to_me, file) for file in files_to_copy])

                msg = _('Application preferences and DAT files restored')
                self.messages_append(f'{msg}.\n{files_copied}', True)
                self.my_logger.info(msg)  # Write to the log file
            except (TypeError, IOError):
                dlg2 = wx.MessageDialog(self,
                                        _('Unable restore the preferences and DAT files from the selected folder.  '
                                        'Check this is the correct folder and check folder & file permissions.'),
                                        _('ERROR'),
                                        wx.OK | wx.ICON_INFORMATION)
                dlg2.ShowModal()
                dlg2.Destroy()
        else:
            dlg2 = wx.MessageDialog(self,
                                    _('Restore process canceled'),
                                    _('INFORMATION'),
                                    wx.OK | wx.ICON_INFORMATION)
            dlg2.ShowModal()
            dlg2.Destroy()
        dlg.Destroy()

    # ------------------------------------------------------------------------------------------------------------------
    # Add a message to the messages control, and include the date/time
    def messages_append(self, message: str, add_delimiter: bool = False) -> None:
        self.controlMessages.AppendText(f'{datetime.datetime.now()}')
        self.controlMessages.AppendText(' - > - ')
        self.controlMessages.AppendText(message)
        if add_delimiter:
            self.controlMessages.AppendText(messageDelimiter)

    # ------------------------------------------------------------------------------------------------------------------
    # Reset the ROMs list column spacing back to default
    def on_reset_spacing(self, _event) -> None:
        self.list.SetColumnWidth(index_exists, exists_col_size)
        self.list.SetColumnWidth(index_rating, rating_col_size)
        self.list.SetColumnWidth(index_number_runs, number_of_runs_col_size)
        self.list.SetColumnWidth(index_favorite, favorite_col_size)
        self.list.SetColumnWidth(index_category, category_col_size)
        self.list.SetColumnWidth(index_system_short_name, short_name_col_size)
        self.list.SetColumnWidth(index_path_media_name, path_media_name_col_size)
        self.list.SetColumnWidth(index_year, year_col_size)
        self.list.SetColumnWidth(index_manufacturer, manufacturer_col_size)
        self.on_resize_window(None)

    # ------------------------------------------------------------------------------------------------------------------
    # This function captures the resizing of the main window and will scale the games list to fill the available space
    def on_resize_window(self, event) -> None:

        self.mc_forced_stop()  # Stop the video player and reset the timer

        # The scaling has to be different if systems support is enabled / disabled
        if self.prefSystemsSupport == 'False':
            adjuster = 200  # Adjust down the horizontal window size as the systems name column is not there
            divider = 2  # Only divide the space change by two columns
        else:
            adjuster = 0
            divider = 3  # Divide the space change by three columns

        default_horizontal_size = 1434 - adjuster

        current_horizontal_size = self.GetSize()[0]

        diff_horizontal_size = current_horizontal_size - default_horizontal_size

        local_rom_name_col_size = self.rom_name_col_size + (diff_horizontal_size // divider)
        local_file_name_col_size = self.file_name_col_size + (diff_horizontal_size // divider)
        local_system_file_name_col_size = self.system_file_name_col_size + (diff_horizontal_size // divider)

        self.list.SetColumnWidth(index_rom_name, local_rom_name_col_size)
        self.list.SetColumnWidth(index_rom_file_name, local_file_name_col_size)
        self.list.SetColumnWidth(index_system_program_file_name, local_system_file_name_col_size)

        if event is not None:
            event.Skip(True)

    # ------------------------------------------------------------------------------------------------------------------
    # This function looks up the short name information for the selected system program in the XML file data.
    # It looks for file size and hash info and compares against the file found.  The comparison is shown as a pop-up.
    def on_xml_info(self, _event) -> None:
        results = self.get_selected_row_rom_file_name()
        rom_file_name = results['rom_file_name']
        rom_name = results['rom_name']
        system_file_name = results['system_file_name']
        short_name = results['short_name']
        if rom_file_name != '':
            xmlinfo.XMLInfo(rom_file_name=rom_file_name,
                            rom_name=rom_name,
                            system_file_name=system_file_name,
                            short_name=short_name,
                            path_executable=self.path_mame_exe,
                            folder_systems=self.folders_systems)

    # ------------------------------------------------------------------------------------------------------------------
    # Copy the default.cfg file to a rom.map file (Where rom is the ROM name selected).  The file is stored
    # in the defined 'Mappings' folder (Which is set in FilePaths).  When the same ROM is later run again, pfeMAME
    # will check to see if the rom.map file exists, and if it does the default.cfg file will be overwritten with it
    # before the ROM is run.  This then allows for per-rom input mapping.
    def on_controller_mapping(self, _event) -> None:
        rom_file_name = self.get_selected_row_rom_file_name()['rom_file_name']

        if self.folder_mapping == '':
            _dlg = wx.MessageDialog(self,
                                    _('You have not set a Mapping folder under File Paths.  This is not a standard MAME '
                                    'folder.  You need to create a folder (Suggest within your MAME folder) called '
                                    'Mapping, and then in the File Paths window you need to set the Input Mapping '
                                    'folder to the folder you have created.'),
                                    _('Mapping Folder Not Defined'),
                                    wx.OK | wx.ICON_INFORMATION
                                    )
            _dlg.ShowModal()
            _dlg.Destroy()
            return
        if self.folder_config == '':
            _dlg = wx.MessageDialog(self,
                                    _('You have not set a Config Folder under file paths.  This needs to be defined so '
                                    'that the default.cfg file can be copied.'),
                                    _('Config Folder Not Defined'),
                                    wx.OK | wx.ICON_INFORMATION
                                    )
            _dlg.ShowModal()
            _dlg.Destroy()
            return
        if not os.path.isfile(os.path.join(self.folder_config, 'default.cfg')):
            _dlg = wx.MessageDialog(self,
                                    _('I am unable to find the default.cfg file within the Config folder set in '
                                    'File Paths.  Check that the folder exists and that within it there is a valid '
                                    'default.cfg file.  If the file is not present, maybe you have not run MAME for '
                                    'the first time so run it and check again if the file exists.'),
                                    _('ERROR - cant find default.cfg'),
                                    wx.OK | wx.ICON_INFORMATION
                                    )
            _dlg.ShowModal()
            _dlg.Destroy()
            return

        # All files and folders seem to be valid so lets proceed.
        # Copy the default.cfg file to the Mapping folder but naming it as rom.map
        try:
            shutil.copyfile(os.path.join(self.folder_config, 'default.cfg'),
                            os.path.join(self.folder_mapping, f'{rom_file_name}.map'))
            self.messages_append(_('Configuration file copied across for ROM file {}\n').format(rom_file_name), True)
            self.my_logger.info(_('Configuration file copied to Mapping folder for ROM file {}').format(rom_file_name))
        except PermissionError:
            _dlg = wx.MessageDialog(self,
                                    _('I am unable to write the {}.map file into the Mapping folder.  '
                                      'Check the folder permissions and that the folder and files within are not '
                                      'Read Only or write protected.  Under Linux check that everyone has permissions '
                                      'to write to the file / folder.').format(rom_file_name),
                                    _('ERROR - cant write map file'),
                                    wx.OK | wx.ICON_INFORMATION
                                    )
            _dlg.ShowModal()
            _dlg.Destroy()
            return

    # ------------------------------------------------------------------------------------------------------------------
    # This function resets any information in a MAME INI file listed against the currently selected program
    def on_ini_file_reset(self, _event) -> None:
        results = self.get_selected_row_rom_file_name()
        rom_file_name = results['rom_file_name']
        rom_name = results['rom_name']
        system_file_name = results['system_file_name']

        # Check that the INI folder is defined
        if self.folder_ini != '' and os.path.exists(self.folder_ini):
            # Check to see if it's a system program.  If it is, we need to delete the system program INI file.
            if system_file_name == '':
                # Treat it as a NON system program
                ini_file_path = os.path.join(self.folder_ini, f'{rom_file_name}.ini')
                if os.path.exists(ini_file_path):
                    os.remove(ini_file_path)
                    self.messages_append(_('INI file for rom {} removed\n').format(rom_name), True)
                else:
                    self.messages_append(_('No INI file for rom {} exists so it cant be removed\n').format(rom_name), True)
            else:
                # Treat it as a SYSTEM program name
                ini_file_path = os.path.join(self.folder_ini,
                                             f'{rom_file_name}_system_{system_file_name}.ini')
                if os.path.exists(ini_file_path):
                    os.remove(ini_file_path)
                    self.messages_append(_('INI file for system program {} removed\n').format(system_file_name), True)
                else:
                    self.messages_append(_('No INI file for system program {}'
                                           ' exists so it cant be removed\n').format(system_file_name), True)

        else:
            self.messages_append(_('Unable to reset the INI file as the INI folder has not been correctly defined\n'),
                                 True)

    # ------------------------------------------------------------------------------------------------------------------
    # The right click has been captured on the games list control so call the popup context menu class
    def on_right_down(self, event) -> None:
        self.PopupMenu(MyPopupMenu(self, event.GetPosition(), self.alternate_emulation_engines))

    # ------------------------------------------------------------------------------------------------------------------
    # Sort the ROMs list control based on the column clicked
    def on_column_clicked(self, _event) -> None:
        # column_selected = functions.get_column_clicked(self.list)  # Get the column that was clicked in
        column_selected = _event.m_col

        # Rebuild the Virtual list with re-ordered data from the self.mylist
        self.mylist.sort(key=lambda x: (x[column_selected].lower(), x[column_selected]))

        # Populate the ROMs list showing all games as default
        self.populate_list(filtercolumn=-1,
                           filtervalue='',
                           read_from_cache=False)
        self.Refresh()
        _event.Skip()

    # ------------------------------------------------------------------------------------------------------------------
    # Show the alternate emulators window
    @staticmethod
    def on_alternate_emulators(_event) -> None:
        alternate_emulators.AlternateEmulators()

    # ------------------------------------------------------------------------------------------------------------------
    # Force pfeMAME to run the selected ROM using MAME rather than the alternate emulator engine assigned
    def on_force_play_mame(self, _event) -> None:
        self.on_double_click(force_mame=True)


# ------------------------------------------------------------------------------------------------------------------
# This is the Class for the right click context menu for the ROMs list
class MyPopupMenu(wx.Menu):
    def __init__(self, parent, position, alternate_emulation_engines) -> None:
        wx.Menu.__init__(self)

        self.parent = parent

        # Only do this once so that we get the bindings done
        if not hasattr(self, 'on_double_click_id'):
            self.on_force_play_mame_id = wx.Window.NewControlId()
            self.on_double_click_id = wx.Window.NewControlId()
            self.on_suppress_short_version_id = wx.Window.NewControlId()
            self.on_verify_roms_id = wx.Window.NewControlId()
            self.on_view_media_types_id = wx.Window.NewControlId()
            self.on_rom_info_id = wx.Window.NewControlId()
            self.on_ini_file_reset_id = wx.Window.NewControlId()
            self.on_add_favorite_id = wx.Window.NewControlId()
            self.on_set_rating_id = wx.Window.NewControlId()
            self.on_xml_info_id = wx.Window.NewControlId()
            self.on_copy_controller_mappping_id = wx.Window.NewControlId()
            self.on_preferences_rom_id = wx.Window.NewControlId()

            # This is used by the help string override function to ensure it's a valid id
            self.context_menu_ids = [self.on_force_play_mame_id,
                                     self.on_double_click_id,
                                     self.on_suppress_short_version_id,
                                     self.on_verify_roms_id,
                                     self.on_view_media_types_id,
                                     self.on_rom_info_id,
                                     self.on_ini_file_reset_id,
                                     self.on_add_favorite_id,
                                     self.on_set_rating_id,
                                     self.on_preferences_rom_id,
                                     self.on_xml_info_id,
                                     self.on_copy_controller_mappping_id
                                     ]

            self.Bind(wx.EVT_MENU, parent.on_force_play_mame, id=self.on_force_play_mame_id)
            self.Bind(wx.EVT_MENU, parent.on_double_click, id=self.on_double_click_id)
            self.Bind(wx.EVT_MENU, parent.on_suppress_short_version, id=self.on_suppress_short_version_id)
            self.Bind(wx.EVT_MENU, parent.on_verify_roms, id=self.on_verify_roms_id)
            self.Bind(wx.EVT_MENU, parent.on_view_media_types, id=self.on_view_media_types_id)
            self.Bind(wx.EVT_MENU, parent.on_rom_info, id=self.on_rom_info_id)
            self.Bind(wx.EVT_MENU, parent.on_ini_file_reset, id=self.on_ini_file_reset_id)
            self.Bind(wx.EVT_MENU, parent.on_add_favorite, id=self.on_add_favorite_id)
            self.Bind(wx.EVT_MENU, parent.on_set_rating, id=self.on_set_rating_id)
            self.Bind(wx.EVT_MENU, parent.preferences_rom_specific, id=self.on_preferences_rom_id)
            self.Bind(wx.EVT_MENU, parent.on_xml_info, id=self.on_xml_info_id)
            self.Bind(wx.EVT_MENU, parent.on_controller_mapping, id=self.on_copy_controller_mappping_id)

        # Build the right click menu
        item, flags = parent.list.HitTest(position)
        if item != -1:  # Don't allow selecting an item on a row with no items
            self.Append(self.on_force_play_mame_id, _('Play ROM with MAME rather than alternate engine'),
                        _('Play the selected ROM with MAME'))
            self.Append(self.on_double_click_id, _('Play ROM'), _('Play the selected ROM'))
            self.Append(self.on_suppress_short_version_id, _('Play ROM without softlist HASH support'),
                        _('Play the selected system program directly without using the softwarelist '
                        'HASH file.  No support for cheats.'))
            self.Append(self.on_copy_controller_mappping_id, _('Copy controller mapping to this ROM'),
                        _('Copies the default.cfg file and associates to this ROM, thus allowing for per-rom input mapping'))
            self.Append(self.on_verify_roms_id, _('Verify ROM'),
                        _('Verify the integrity of the selected ROM'))
            self.Append(self.on_view_media_types_id, _('View Media Types'),
                        _('View the types of media that this system is capable of running'))
            self.Append(self.on_rom_info_id, _('ROM information'),
                        _('Get information on the selected ROM'))
            self.Append(self.on_xml_info_id, _('Analyse XML info'),
                        _('Check XMl file information for this system program'))
            self.Append(self.on_ini_file_reset_id, _('Reset INI File'),
                        _('Reset the INI configuration file for this program'))
            self.Append(self.on_add_favorite_id, _('Add to favorites'),
                        _('Add the selected ROM to my favorites list'))
            self.Append(self.on_set_rating_id, _('ROM Rating'),
                        _('Set the star rating for the selected ROM'))
            self.Append(self.on_preferences_rom_id, _('ROM Preferences'),
                        _('Set the MAME preferences for this ROM'))

            self.Bind(wx.EVT_MENU_HIGHLIGHT, self.on_help_string)
            self.Bind(wx.EVT_MENU_CLOSE, self.on_remove_text)

            # Get the previous status bar message for safe keeping - it will be restored at the end
            temp = parent.statusbar_panel.get_status_text()

            parent.list.Select(item)  # make the ROM we right clicked selected

            results = parent.get_selected_row_rom_file_name()

            # Try to get the file short name (for systems). If it doesn't exist then disable some menu items.
            if results['short_name'] == '':
                self.Enable(id=self.on_suppress_short_version_id, enable=False)

            # Enable / disable the 'force play with mame' option as we should only show this if there is an option
            # available to play with a different emulation engine.
            if not any(results['rom_file_name'] in nested_list for nested_list in alternate_emulation_engines):
                self.Enable(id=self.on_force_play_mame_id, enable=False)

            parent.statusbar_panel.set_status(temp)

    # ------------------------------------------------------------------------------------------------------------------
    # Display the help string in the status bar for the ROM list right-click menu
    #
    # By default, wxPython should be doing this for me, but there seems to be an issue with the current release
    # whereby the toolbar menu works ok but the right-click context menu won't display help text.
    #
    # The check for the event_id being from the list of valid ids for the context menu is there to prevent a core
    # error event.  This happens when you try to get an argument that doesn't exist.
    def on_help_string(self, event) -> None:
        event_id = event.GetId()
        if event_id in self.context_menu_ids:
            self.parent.statusbar_panel.set_status(self.GetHelpString(event_id))

    # ------------------------------------------------------------------------------------------------------------------
    # We are no longer selecting any items in the context menu so reset the status bar to 'Ready'
    def on_remove_text(self, _event) -> None:
        self.parent.statusbar_panel.set_status(_('Ready'))


# ------------------------------------------------------------------------------------------------------------------
# Custom renderer for the UltimateListCtrl header drawing so that we can have custom colors
class UltimateHeaderRenderer(object):
    __slots__ = ['_hover', '_pressed', 'application_colors']

    def __init__(self, application_colors: dict):
        self._hover = False
        self._pressed = False
        self.application_colors = application_colors
        
    def DrawHeaderButton(self, dc, rect, flags):
        self._hover = False
        self._pressed = False
        
        color = self.application_colors['control_bar']

        dc.SetBrush(wx.Brush(color, wx.SOLID))
        dc.SetBackgroundMode(wx.SOLID)
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(rect)
        dc.SetBackgroundMode(wx.TRANSPARENT)

    def GetForegroundColour(self):
        pass
        """
        if self._hover:
            return wx.Colour(255,255,255)
        else:
            return wx.Colour(0,0,0)
        """


# ------------------------------------------------------------------------------------------------------------------
class SearchCtrl2(wx.TextCtrl):
    """
     This is the custom search control so that I can set the foreground and background colors which you
    cannot do with the standard control across all operating systems.  It also gets rid of the usual hint text and
    instead just inserts a 'Search' text so that I can control the colour.
    """
    def __init__(self,
                 parent: wx.Panel,
                 id: int = wx.ID_ANY,
                 value: str = _('Search'),
                 size: tuple = wx.DefaultSize,
                 style: int = 0,
                 validator=wx.DefaultValidator,
                 name: str = wx.TextCtrlNameStr):

        wx.TextCtrl.__init__(self, parent=parent, id=id, value=value, size=size, style=style,
                             validator=validator, name=name)

        self.Bind(wx.EVT_KILL_FOCUS, self.on_lost_focus)
        self.Bind(wx.EVT_TEXT, self.on_set_text)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_clear_field)

        self.override = False

    def on_lost_focus(self, event):
        self.set_default_text()

    def on_set_text(self, _event) -> None:
        if not self.override:
            if self.GetValue() == '':
                # Use callafter to prevent a circular event issue
                wx.CallAfter(self.set_default_text)
            # If we have deleted search text and then start typing again, strip off the 'Search' hint
            # as that will mess up the search
            elif self.GetValue().endswith(_('Search')):
                self.override = True
                temp_value = self.GetValue().strip(_('Search'))
                self.Clear()
                self.AppendText(temp_value)
                self.override = False

    def set_default_text(self) -> None:
        self.override = True
        self.SetValue(_('Search'))
        self.override = False

    def clear_default_text(self) -> None:
        self.override = True  # Stop the set text function from putting something back in
        self.Clear()
        self.SetFocus()
        self.override = False

    def on_clear_field(self, _event) -> None:
        # Use callafter to prevent a circular event issue
        if self.GetValue() == _('Search'):
            wx.CallAfter(self.clear_default_text)


# ------------------------------------------------------------------------------------------------------------------
class ImagePanel(wx.Panel):
    """
    This is the class for my expanding image panel.  The panel will expand to fit the image in the centre of the
    space available and use padding around the image to centre it.
    """
    def __init__(self, parent: wx.Panel, image_x_size: int, image_y_size: int, rom_list):

        wx.Panel.__init__(self, parent)

        self.SetMinSize((image_x_size, image_y_size))  # Make sure it can never get smaller than this

        self.parent = parent
        self.rom_list = rom_list
        self.image = wx.StaticBitmap(self, bitmap=wx.Bitmap(image_x_size, image_y_size))
        self.image.Bind(wx.EVT_LEFT_DOWN, self.on_clicked)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)

        self.stretch_size = (1, 1)

        sizer.Add(self.stretch_size, proportion=1, flag=wx.EXPAND)
        sizer2.Add(self.stretch_size, proportion=1, flag=wx.EXPAND)
        sizer2.Add(self.image)
        sizer2.Add(self.stretch_size, proportion=1, flag=wx.EXPAND)
        sizer.Add(sizer2, proportion=0, flag=wx.EXPAND)
        sizer.Add(self.stretch_size, proportion=1, flag=wx.EXPAND)
        self.SetSizer(sizer)

    # Show the pop-up large image viewer if we have clicked in this image
    def on_clicked(self, _event) -> None:
        current_item = int(self.rom_list.GetFirstSelected())
        image_file_name = MainFrame.image_file_name
        if current_item != -1 and os.path.exists(image_file_name):
            rom_name = self.rom_list.GetItem(current_item, index_rom_name).GetText()
            popupimage.PopUpImageViewer(rom_name, image_file_name)


# ------------------------------------------------------------------------------------------------------------------
# This is the class for my virtual list control
class VirtualList(ulc.UltimateListCtrl):
    __slots__ = ['parent', 'pref_icon_size', 'pref_show_rom_icons', 'folders_icons', 'pref_highlight_missing',
                 'application_colors', 'my_logger']

    def __init__(self,
                 parent: wx.Panel,
                 pref_icon_size: str,
                 pref_show_rom_icons: str,
                 folders_icons: str,
                 pref_highlight_missing: str,
                 application_colors: dict,
                 my_logger):

        ulc.UltimateListCtrl.__init__(self,
                                      parent, 
                                      -1, 
                                      size=(1041, 400),
                                      agwStyle=wx.LC_REPORT 
                                             | wx.LC_VIRTUAL 
                                             | wx.LC_VRULES 
                                             | wx.LC_HRULES 
                                             | wx.LC_SINGLE_SEL 
                                             | ulc.ULC_USER_ROW_HEIGHT)

        #  These are needed for some functions below
        self.pref_icon_size = pref_icon_size
        self.pref_show_rom_icons = pref_show_rom_icons
        self.folders_icons = folders_icons
        self.pref_highlight_missing = pref_highlight_missing
        self.my_logger = my_logger
        self.parent = parent

        # Create the columns
        bold_font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        bold_font.SetWeight(wx.BOLD)
        bold_font.SetPointSize(8)
        info = ulc.UltimateListItem()
        info._mask = wx.LIST_MASK_TEXT | wx.LIST_MASK_WIDTH | wx.LIST_MASK_FORMAT | ulc.ULC_MASK_FONT

        def add_one_column(column: int, text: str, width: int, font:wx.Font, position_format) -> None:
            info._text = text
            info._width = width
            info._font = font
            info._format = position_format
            custom_render = UltimateHeaderRenderer(application_colors)
            info.SetCustomRenderer(custom_render) 
            self.InsertColumnInfo(column, info)


        columns_list = [
            (
                index_rom_name,
                _('ROM Name'),
                rom_name_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_rom_file_name,
                _('File Name'),
                file_name_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_system_program_file_name,
                _('System File'),
                system_file_name_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_system_short_name,
                _('Short Name'),
                short_name_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_exists,
                _('Exists'),
                exists_col_size,
                bold_font,
                ulc.ULC_FORMAT_CENTER
            ),
            (
                index_rating,
                _('Rating'),
                rating_col_size,
                bold_font,
                ulc.ULC_FORMAT_CENTER
            ),
            (
                index_number_runs,
                _('Runs'),
                number_of_runs_col_size,
                bold_font,
                ulc.ULC_FORMAT_CENTER
            ),
            (
                index_favorite,
                _('Fav'),
                favorite_col_size,
                bold_font,
                ulc.ULC_FORMAT_CENTER
            ),
            (
                index_category,
                _('Category'),
                category_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_path_media_name,
                _('Path Media Name'),
                path_media_name_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_year,
                _('Year'),
                year_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            ),
            (
                index_manufacturer,
                _('Manufacturer'),
                manufacturer_col_size,
                bold_font,
                ulc.ULC_FORMAT_LEFT
            )
        ]
        for line in columns_list:
            add_one_column(*line)

        # Build the imagelist - Used for all image graphics in the control
        # Image 0 is blank
        # Image 1 is the game image 
        #   (changed on the fly when drawing so that we use a 'virtual' image list which is a lot faster)
        # Image 2 ~ 7 are the star rating images
        # Image 8 - 9 are the tick and cross
        self.my_logger.info(_('Started building icon list.'))  # Write to the log file
        self.icon_list = ulc.PyImageList(width=15, height=15)
        # Create a transparent scaled blank icon
        bitmap = wx.Bitmap.FromRGBA(int(pref_icon_size), int(pref_icon_size), alpha=0)
        self.icon_list.Add(bitmap)  # Image 0 - always blank
        self.icon_list.Add(bitmap)  # Image 1 - dynamically updated game icon image
        self.icon_list.Add(bitmap)  # Image 2 - Zero star rating image

        images = ['stars_1.gif',
                  'stars_2.gif',
                  'stars_3.gif',
                  'stars_4.gif',
                  'stars_5.gif',
                  'cross.png',
                  'tick.png']
        for item in images:
            self.icon_list.Add(wx.Bitmap(os.path.join(path_to_me, 'images', item)))

        self.SetImageList(self.icon_list, wx.IMAGE_LIST_SMALL)

        self.SetUserLineHeight(int(pref_icon_size))  # Set each row height to match the icon image scale size

        self.my_logger.info(_('Completed building icon list.'))  # Write to the log file

        # Set the number of items
        self.SetItemCount(len(MainFrame.mylist_filtered))

        # Set the attributes for alternating row colors
        self.attr1 = ulc.UltimateListItemAttr()
        self.attr1.SetBackgroundColour(wx.Colour(application_colors['row_alt_2']))
        self.attr2 = ulc.UltimateListItemAttr()
        self.attr2.SetBackgroundColour(wx.Colour(application_colors['row_alt_1']))

        # Set the attribute for the missing game highlight
        self.attr3 = ulc.UltimateListItemAttr()
        self.attr3.SetBackgroundColour(wx.Colour(application_colors['missing_game']))

    # ------------------------------------------------------------------------------------------------------------------
    # The following are the callbacks for the virtual list control

    # When requested, provide the list control the text to be displayed in row(item) / col
    def OnGetItemText(self, item: int, col: int) -> str:
        if col is not index_rating \
                and col is not index_exists \
                and col is not index_favorite \
                and (0 <= item < self.GetItemCount()):
            return MainFrame.mylist_filtered[item][col]
        else:
            # Don't show the star rating number, exists flag, and favorite flag (An image will be inserted later)
            return ''

    # Not used as we need the function below instead
    def OnGetItemImage(self, item) -> int:
        return -1

    # Upon request, provide the list control the object to be displayed in row / col.
    # Some rows are omitted given that we are displaying icons (e.g. the 5-stars rating)
    # This is implemented as a 'virtual' image list in that we replace one image on the fly
    # and use that image for the game - This is very fast compared to the normal method
    # of pre-building the full image list.
    def OnGetItemColumnImage(self, item: int, column: int = 0) -> Union[int, list]:
        # This function allows a bitmap to be scaled
        def scale_bitmap(bitmap_local: wx.Bitmap, width: int, height: int) -> wx.Bitmap:
            image = wx.Bitmap.ConvertToImage(bitmap_local).Rescale(width, height, wx.IMAGE_QUALITY_NORMAL)
            result = wx.Bitmap(image)
            return result

        if column == index_rom_name \
                and self.pref_show_rom_icons == 'True' \
                and 0 <= item < self.GetItemCount():
            # Use the ROM icons
            rom_file_name = MainFrame.mylist_filtered[item][index_rom_file_name]
            try:
                if os.path.isfile(bitmap_path := os.path.join(self.folders_icons, f'{rom_file_name}.ico')):
                    bitmap = wx.Bitmap(bitmap_path)
                    # Scale the ROM icons
                    bitmap = scale_bitmap(bitmap, int(self.pref_icon_size), int(self.pref_icon_size))
                    self.icon_list.Replace(1, bitmap)
                else:
                    # We are unable to return an icon image so just return a blank one
                    return 0
            except IndentationError:
                return 0
            return 1
        elif column == index_rating and 0 <= item < self.GetItemCount():  # Use the ratings icons
            return [
                int(MainFrame.mylist_filtered[item][index_rating]) + 2]  # Return the stars section from the imagelist
        elif column == index_exists and 0 <= item < self.GetItemCount():
            #  Show the exists column ticks and crosses
            return int(MainFrame.mylist_filtered[item][index_exists]) + 8
        elif column == index_favorite and 0 <= item < self.GetItemCount():
            #  Show the Favorites column ticks & crosses
            return int(MainFrame.mylist_filtered[item][index_favorite]) + 8
        else:
            # We got here because we are in a column where no icon image should be shown.
            # Just return and empty result so no image is displayed.
            return []

    # Alternate the row colors and set missing ROMs as a red highlight
    def OnGetItemAttr(self, item: int):
        if item < 0:
            return None
        if MainFrame.mylist_filtered[item][index_exists] == '0' and self.pref_highlight_missing == 'True':
            return self.attr3  # Return the missing color
        if item % 2:
            return self.attr1
        else:
            return self.attr2

    def OnGetItemColumnCheck(self, item, column=0):
        return []

    def OnGetItemColumnKind(self, item, column=0):
        return -1

    def OnGetItemKind(self, item):
        return 0

    def OnGetItemTextColour(self, item, col):
        return None

    def OnGetItemToolTip(self, item, col):
        return ''

    # This override is to fix a really annoying issue that only pops up in Linux.
    # In some cases you only have to move the mouse up and over the ULC header,
    # and in other cases you have to try and drag over the header.  In either
    # case it raises an exception due to an out of range index.  The try/except
    # below is all that is needed to fix it.
    def GetItem(self, itemOrId, col: int = 0) -> str:
        """
        Returns the information about the input item.

        :param itemOrId: an instance of :class:`UltimateListItem` or an integer specifying
         the item index;
        :param col: the column to which the item belongs to.
        """

        item = ulc.CreateListItem(itemOrId, col)
        try:
            return self._mainWin.GetItem(item, col)
        except Exception:
            return ''


# ----------------------------------------------------------------------------------------------------------------------
# This is where the application starts
if __name__ == '__main__':
    app = wx.App()

    # Define Global variable values here
    path_to_me = functions.path_to_me()  # Get the absolute path to me
    default_image_file_name = os.path.join(path_to_me, 'images', 'default.gif')

    # import gettext
    # gettext.install('app')

    # _ = functions.set_up_language()  # Set up the application language

    MainFrame()
    app.MainLoop()
